#!/bin/bash

# Check if package version number specified
# If not, use default version number
if [ -z "$1" ]
then 
    echo "WARNING: PACKAGE_VERSION not given, using 0.3.3"
    export PACKAGE_VERSION=0.3.3
else
    export PACKAGE_VERSION=$1
fi

PACKAGE_NAME=ti-openmpacc_${PACKAGE_VERSION}
CLACC_DIR=clacc
CLACC_ARM=clacc_arm
CLACC_x86=clacc_x86
LIBOMPACC=libOpenMPAcc.so.${PACKAGE_VERSION}
BUILD_DIR=build
STAGING_DIR=staging

# Check build platform machine arch to figure out if 
# build is cross or not
BUILD_MACHINE="$(uname -m)"
if [[ ${BUILD_MACHINE} != *arm* ]]
then
    export CROSS_BUILD=1
    echo "************************************"
    echo "x86* platform: CROSS BUILD triggered"
    echo "************************************"
else
    echo "************************************"
    echo "ARM platform: NATIVE BUILD triggered"
    echo "************************************"
fi

# Only build boost in cross compile mode
# In native mode, we assume that the required boost 
# libs are apt-get installed
if [ ${CROSS_BUILD} ]
then 
    # Options for dowloading and building boost
    BOOST_INSTALL_DIR=boost
    BOOST_BUILD_SCRIPT=build_boost.sh
    export BOOST_VERSION=1_48_0
    export BOOST_BUILD_LIBS=program_options,filesystem,system
    export BOOST_ROOT_ARM=$(pwd)/${BOOST_INSTALL_DIR}/boost_arm
    export BOOST_ROOT_X86=$(pwd)/${BOOST_INSTALL_DIR}/boost_x86
    # Build and install boost if not done already
    echo "************************************"
    echo "******* Building Boost *************"
    echo "************************************"
    pushd ${BOOST_INSTALL_DIR}
    ./${BOOST_BUILD_SCRIPT}
    popd
    echo "************************************"
    echo "******* Built Boost ****************"
    echo "************************************"
fi
# Make clacc
echo "************************************"
echo "******* Making Clacc ***************"
echo "************************************"
make -C ${CLACC_DIR} clean
make -C ${CLACC_DIR}
# Check if clacc executables were built
if [ ! -e ${CLACC_DIR}/${CLACC_ARM} ]
then
    echo ">> ERROR in making ${CLACC_ARM}"
    exit 1
fi
if [ ${CROSS_BUILD} ]
then
    if [ ! -e ${CLACC_DIR}/${CLACC_X86} ]
    then
        echo ">> ERROR in making ${CLACC_X86}"
        exit 1
    fi
fi
echo "************************************"
echo "******* Clacc Ready ****************"
echo "************************************"

# Make OpenMPAcc package
echo "************************************"
echo "******* Making libOpenMPAcc ********" 
echo "************************************"
rm -rf ${BUILD_DIR}
rm -rf ${STAGING_DIR}/${PACKAGE_NAME}
mkdir ${BUILD_DIR}
pushd ${BUILD_DIR}
if [ ${CROSS_BUILD} ]
then
    cmake -DCMAKE_TOOLCHAIN_FILE=../CMakeARMToolChain.txt -DCROSS_BUILD=ON ..
else
    cmake ..
fi
make install
popd
# Check if package install directory exists 
# with libOpenMPAcc.so
if [ ! -e ${STAGING_DIR}/${PACKAGE_NAME}/lib/${LIBOMPACC} ]
then
    echo ">> ERROR in making ${LIBOMPACC}"
    exit 1
fi
echo "************************************"
echo "******* libOpenMPAcc Ready *********"
echo "************************************"
