/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file clacc.cc
 * \brief Compiler for OpenMP 4.0 Accelerator Model programs 
 */

#define CLACC_USE_ONLY
#include "clacc.h"

/*********************************************************************/
/* Main                                                              */
/*********************************************************************/
int main(int argc, char* argv[])
{
    /*********************************************************************/
    /* Setup option parser and and print help if required                */
    /*********************************************************************/
    po::options_description help_args = SetupCommandLineParser(argc, argv);

    /*********************************************************************/
    /* Initialize compiler options                                       */
    /*********************************************************************/
    InitCompilerOpts(help_args);

    /*********************************************************************/
    /* Handle all command line options                                   */
    /*********************************************************************/
    HandleCommandLineOpts();
    
    std::vector<std::string> input_c_files;
    std::vector<std::string> host_int_files;
    std::vector<std::string> clocl_cl_chain_list;
    std::vector<std::string> generated_tmp_files;

    try
    {
        /*********************************************************************/
        /* Check input file existence and separate object files              */
        /*********************************************************************/
        CheckAndSeparateInputFiles(input_c_files);

        /*********************************************************************/
        /* Initialize compilers                                              */
        /*********************************************************************/
        Compiler h_cc(host_cc, HOST ,debug);
        Compiler h_link_cc(host_link_cc, HOST ,debug);
        Compiler t_cc(target_cc, TARGET ,debug);
        Compiler s2s(target_cc, S2S ,debug);
        Compiler clocl(cl_link_cc, TARGET ,debug);

        /*********************************************************************/
        /* Set compiler options                                              */
        /*********************************************************************/
        h_cc.set_compile_opts(host_cc_opts);
        h_link_cc.set_compile_opts(host_cc_opts);
        h_link_cc.set_link_lib_dependencies(host_cc_libs);
        t_cc.set_compile_opts(target_cc_opts);
        s2s.set_compile_opts(s2s_opts);
        clocl.set_compile_opts(clocl_opts);

        /*********************************************************************/
        /* Target compile stage                                              */
        /*********************************************************************/
        MakeTargetObjs(t_cc, 
                       input_c_files, 
                       generated_tmp_files
                       );

        /*********************************************************************/
        /* S2S lowering stage                                                */
        /*********************************************************************/
        DoS2SLowering(s2s,
                      input_c_files,
                      clocl_cl_chain_list, 
                      host_int_files,
                      generated_tmp_files
                      );
        
        /*********************************************************************/
        /* CLOCL stage                                                       */
        /*********************************************************************/
        MakeCloclArtifacts(clocl,
                           h_link_cc,
                           clocl_cl_chain_list,
                           generated_tmp_files
                           );

        /*********************************************************************/
        /* Final link stage                                                  */
        /*********************************************************************/
        DoFinalLink(h_cc,
                    h_link_cc,
                    input_c_files,
                    host_int_files,
                    generated_tmp_files
                    );

        /*********************************************************************/
        /* Clean-up stage                                                    */
        /*********************************************************************/
        CleanUp(generated_tmp_files);

    }
    catch(fs::filesystem_error& e)
    {
        std::cout << ">> ERROR: Filesystem: " << e.what() << std::endl;
        return 1;
    }
    catch(std::exception& e)
    {
        std::cout << e.what() << std::endl;
        return 1;
    }

    return 0;
}

/*********************************************************************/
/* Print Usage                                                       */
/*********************************************************************/
void PrintUsage(po::options_description& visible)
{
    std::cout << "TI CLACC OpenMP Accelerator Model Compiler" 
              << std::endl
              << "Version:    " << clacc_version 
              << std::endl
#ifdef BUILD_DATE
              << "Build Date: " << build_date 
              << std::endl
#endif
              << "Usage:      clacc [options] [object-files] C-source-files" 
              << std::endl 
              << "Note:       Environment Variables " 
              << ti_ocl_cgt_env_var << "(TI CGT Install Location), " 
              << ti_rootdir_env_var << "(TI MCSDK Root Filesystem Location) must be set" 
              << std::endl
              << visible 
              << std::endl;
}

/*********************************************************************/
/* Check and get required environment variables                      */
/*********************************************************************/
void GetEnvVars(po::options_description& help_args)
{
    /* Get TI_OCL_CGT_INSTALL value, and exit if it is NULL */
    char* ti_ocl_cgt_install_path =  getenv(ti_ocl_cgt_env_var.c_str());
    if (ti_ocl_cgt_install_path == NULL)
    {
        PrintUsage(help_args);
        std::cerr << ti_ocl_cgt_env_var
                  << " not defined, exiting.." 
                  << std::endl;
        exit(EXIT_FAILURE);
    }
    ti_ocl_cgt_install = std::string(ti_ocl_cgt_install_path);

    /* Get TARGET_ROOTDIR value, but do not exit if it is NULL */
    char* ti_rootdir_path_cstr =  getenv(ti_rootdir_env_var.c_str());
    if (ti_rootdir_path_cstr == NULL)
    {
        ti_rootdir_path = "";
    }else
    {
        ti_rootdir_path = std::string(ti_rootdir_path_cstr);
    }
}

/*********************************************************************/
/* Initialize Compiler Options for all compilers being used          */
/*********************************************************************/
void InitCompilerOpts(po::options_description& help_args)
{
    /* Check and get cgt and target rootdir paths */
    GetEnvVars(help_args);

    /* Include path for cl6x */
    std::string cgt_include_path = "-I";
    cgt_include_path.append(ti_ocl_cgt_install);
    cgt_include_path.append("/include");

    /* Lib path for cl6x */
    std::string cgt_lib_path = "-L";
    cgt_lib_path.append(ti_ocl_cgt_install);
    cgt_lib_path.append("/lib");

    /* Lib path for g++/gcc. OpenCL, OpenMPAcc libs are present here */
    std::string ti_rootdir_lib_path = ti_rootdir_path + "/usr/lib";

    /* Include path for cl6x for DSP side omp.h */
    std::string ti_rootdir_dsp_inc_path = ti_rootdir_path 
                                          + "/usr/share/ti/openmpacc/dsp/include";

    /* Lib path prefix for g++/gcc that allows finds OpenCL depency 
     * libs such as cmem, hyplink, bfd, sqlite3 etc*/
    std::string ti_rootdir_rpath_prefix = "-Wl,-rpath-link,";

    clocl_opts = {"-t"};
    
    host_cc_opts = {"-O3", 
                    "-Wall", 
                    "-Wextra", 
                    "-fopenmp"
                    };

    host_cc_libs = {"-L" + ti_rootdir_lib_path,  
                    "-lOpenMPAcc", 
                    "-lOpenCL", 
                    "-locl_util",
                    ti_rootdir_rpath_prefix + ti_rootdir_lib_path 
                    };

    target_cc_opts_fixed = {"--openmp:device", 
                            "--abi=elfabi", 
                            "-mv6600", 
                            cgt_include_path,
                            "-I" + ti_rootdir_dsp_inc_path
                            };

    target_cc_opts_variable = {"-O3"};

    s2s_opts_fixed = {"--openmp:host", 
                      "--abi=elfabi", 
                      "--cbe", 
                      cgt_include_path,
                      "-I" + ti_rootdir_dsp_inc_path
                      };
}

/*********************************************************************/
/* Setup Command Line parser and print help if required              */
/*********************************************************************/
po::options_description SetupCommandLineParser(int argc, char* argv[])
{
    /*********************************************************************/
    /* Declare command line options visible to user                      */
    /*********************************************************************/
    po::options_description visible("Options");
    visible.add_options()
        ("help,h","Show help message")
        ("verbose,v","Show debug output")
        ("keep_files,k", "Keep intermediate temporary files")
        ("runtime_debug,d", "Enable runtime debugging information display")
        ("runtime_performance,p", "Enable runtime performance information display")
        ("lib,l", po::value< std::vector<std::string> >()->composing(),
         "Add library for host side linking")
        ("lib_path,L", po::value< std::vector<std::string> >()->composing(),
         "Add dir to searched host library directories")
        ("inc_path,I", po::value< std::vector<std::string> >()->composing(),
         "Add dir to searched host include directories")
        ("show_cc_opts", "Show host and target compiler options being used")
        ("host_cc", po::value<std::string>(),
         "Host compiler to use")
        ("host_cc_opts", po::value< std::vector<std::string> >(),
         "Host compiler options to override default options")
        ("target_cc_opts", po::value< std::vector<std::string> >(),
         "Target compiler options to override default options")
        ("exe,o", po::value<std::string>(), 
         "Name of executable fat binary")
        ;

    /*********************************************************************/
    /* Allowed hidden options, not visible to user                       */
    /*********************************************************************/
    po::options_description hidden("Hidden options");
    hidden.add_options()
        ("input-file", po::value< std::vector<std::string> >(), "input file")
        ("host_objs", po::value< std::vector<std::string> >(),
         "Pre-compiled host object files to link")
        ("target_objs", po::value< std::vector<std::string> >(),
         "Pre-compiled target object files to link")
        ;
    
    try
    {
        po::options_description cmdline_options;
        cmdline_options.add(visible).add(hidden);

        po::positional_options_description p;
        p.add("input-file", -1);

        po::parsed_options parsed = po::command_line_parser(argc, argv)
            .options(cmdline_options)
            .positional(p)
            .run();

        po::store(parsed, vm);
        po::notify(vm);

        /* Check if help required */
        if (vm.count("help")) 
        {
            PrintUsage(visible);
            exit(EXIT_SUCCESS);
        }
    }
    catch(...)
    {
        PrintUsage(visible);
        exit(EXIT_FAILURE);
    }

    return visible;
}

/*********************************************************************/
/* Handle command line options                                       */
/*********************************************************************/
void HandleCommandLineOpts()
{
    std::vector<std::string> libs;
    std::vector<std::string> lib_paths;
    std::vector<std::string> inc_paths;
    std::vector<std::string> host_cc_libs_user;

    if (vm.count("verbose")) 
    {
        debug = true;
        clocl_opts.push_back("-v");
    }

    if (vm.count("keep_files")) 
    {
        keep_tmp_files = true;
    }
    
    if (vm.count("runtime_debug")) 
    {
        runtime_debug = true;
    }
    
    if (vm.count("runtime_performance")) 
    {
        runtime_performance = true;
    }
    
    if (vm.count("host_cc"))
    {
        host_cc = vm["host_cc"].as<std::string>();
        if (debug)
        {
            std::cout << ">> DEBUG: Using host compiler: " 
                      << host_cc 
                      << std::endl;
        }
    }

    if (vm.count("host_cc_opts"))
    {
        host_cc_opts = vm["host_cc_opts"]
                       .as< std::vector<std::string> >(); 
        if (debug)
        {
            std::cout << ">> DEBUG: Override host cc opts: " 
                      << host_cc_opts 
                      << std::endl;
        }
    }

    if (vm.count("target_cc_opts"))
    {
        target_cc_opts_variable = vm["target_cc_opts"]
                                  .as< std::vector<std::string> >(); 
        s2s_opts_variable = vm["target_cc_opts"]
                            .as< std::vector<std::string> >();
        if (debug)
        {
            std::cout << ">> DEBUG: Override target cc opts: " 
                      << target_cc_opts_variable 
                      << std::endl;
        }
    }

    if (vm.count("lib"))
    {
        libs = vm["lib"].as< std::vector<std::string> >();
        if (debug)
        {
            std::cout << ">> DEBUG: Host side link libraries: " 
                      << libs
                      << std::endl;
        }
    }

    if (vm.count("lib_path"))
    {
        lib_paths = vm["lib_path"].as< std::vector<std::string> >();
        
        if (debug)
        {
            std::cout << ">> DEBUG: Library path added: " 
                      << lib_paths 
                      << std::endl;
        }
    }

    if (vm.count("inc_path"))
    {
        inc_paths = vm["inc_path"].as< std::vector<std::string> >();
   
        if (debug)
        {
            std::cout << ">> DEBUG: Include path added: " 
                      << inc_paths 
                      << std::endl;
        }
    }
    /* Add the user specified -L paths*/
    for (auto path: lib_paths)
    {
        host_cc_libs_user.push_back("-L" + path);
    }
    /* Add the user specified -l libs*/
    for (auto lib: libs)
    {
        host_cc_libs_user.push_back("-l" + lib);
    }
    /* Add the user specified -I paths in order */
    for (auto path: inc_paths)
    {
        host_cc_opts.push_back("-I" + path);
    }

    /* Compose host_cc_libs */
    /* Add user lib paths and libs into the front of host_cc_libs */
    if (host_cc_libs_user.size()!=0)
    {
        std::vector<std::string>::reverse_iterator rit;
        /* Take the items of host_cc_libs_user in reverse order and 
         * insert them into host_cc_libs to ensure consistency in ordering */
        /* TODO: Use std::dequeue for host_cc_libs in next release 
         * for efficient front insertion in this instance */
        for (rit = host_cc_libs_user.rbegin(); 
             rit != host_cc_libs_user.rend(); 
             ++rit)
        {
            auto it = host_cc_libs.begin();
            host_cc_libs.insert(it,*rit);
        }
    }

    /* Create target_cc_opts and s2s_opts as variable opts have 
     * now been specified */
    /* Add fixed opts */
    target_cc_opts = target_cc_opts_fixed;
    s2s_opts = s2s_opts_fixed;
    /* Add variable opts */
    target_cc_opts.insert(target_cc_opts.end(), 
                          target_cc_opts_variable.begin(),
                          target_cc_opts_variable.end()
                          );
    /* S2S variable opts may not have been set, so check for it first*/
    if (s2s_opts_variable.size()!=0)
    {
        s2s_opts.insert(s2s_opts.end(), 
                        s2s_opts_variable.begin(),
                        s2s_opts_variable.end()
                        );
    }
    
    if (vm.count("show_cc_opts")) 
    {
        std::cout << ">> Host CC: " << host_cc
                  << " " << host_cc_opts 
                  << std::endl;
        std::cout << ">> Target CC: " << target_cc 
                  << " " << target_cc_opts 
                  << std::endl;
    }
   

    if (vm.count("exe"))
    {
        exe_name = vm["exe"].as<std::string>();
        if (debug)
        {
            std::cout << ">> DEBUG: Executable fat binary name: " 
                      << exe_name 
                      << std::endl;
        }
    }

    if (vm.count("host_objs"))
    {
        user_host_objs = vm["host_objs"].as< std::vector<std::string> >(); 
        host_objs = user_host_objs;
        if (debug)
        {
            std::cout << ">> DEBUG: Pre-compiled host objs: " 
                      << host_objs 
                      << std::endl;
        }
    }

    if (vm.count("target_objs"))
    {
        target_objs = vm["target_objs"].as< std::vector<std::string> >(); 
        if (debug)
        {
            std::cout << ">> DEBUG: Pre-compiled target objs: " 
                      << target_objs 
                      << std::endl;
        }
    }

    if (vm.count("input-file"))
    {
        input_files = vm["input-file"].as< std::vector<std::string> >();
        if (debug)
        {
            std::cout << ">> DEBUG: Input files: " << input_files 
                      << std::endl;
        }
    }else
    {
        std::cerr << ">> ERROR: No input files" 
                  << std::endl;
        exit(EXIT_FAILURE);
    }
    
}

/*********************************************************************/
/* Check each given input file for existence and separate them       */
/* into lots of C source files and ARM host and C66 target           */
/* object files                                                      */
/*********************************************************************/
void CheckAndSeparateInputFiles(std::vector<std::string>& input_c_files)
{
    ElfInit();
    for (auto infile : input_files)
    {
        if(!fs::exists(fs::path(infile)))
        {
            std::cerr << ">> ERROR: File " << infile 
                      << " not found" 
                      << std::endl;
            exit(EXIT_FAILURE);
        }    

        /* If it is a source code file */
        if(!IsElf(infile))
        {
            if (debug)
            {
                std::cout << ">> DEBUG: Source code file: " << infile 
                          << std::endl;
            }
            input_c_files.push_back(infile);
        }
        else
        {
            /* It is an ELF object file */
            switch(GetElfMachineID(infile))
            {
                case EM_ARM:
                    {
                        if (debug)
                        {
                            std::cout << ">> DEBUG: "
                                      << "ARM object file or archive: " 
                                      << infile 
                                      << std::endl;
                        }
                        host_objs.push_back(infile);
                        user_host_objs.push_back(infile);
                        break;
                    }
                case EM_TI_C6000:
                    {
                        if (debug)
                        {
                            std::cout << ">> DEBUG: "
                                      << "TI C66 DSP object file or archive: " 
                                      << infile 
                                      << std::endl;
                        }
                        target_objs.push_back(infile);
                        break;
                    }
                default:
                    {
                        std::cerr << ">> ERROR: Unsupported ELF Object " 
                                  << infile 
                                  << std::endl; 
                        exit(EXIT_FAILURE);
                    }
            }
        }
    }
}

/*********************************************************************/
/* Create object files for target                                    */
/*********************************************************************/
void MakeTargetObjs(Compiler& t_cc,
                    std::vector<std::string>& input_c_files,
                    std::vector<std::string>& tmp_list)
{
    /* Create .obj files  */
    std::string obj_file;
    for(auto infile : input_c_files)
    {
        /* Generate .obj files for each input file */
        t_cc.Compile(infile);
        obj_file = fs::path(infile).stem().string();
        /* Check existence and add the generated obj file name 
         * to target obj list 
         * */
        obj_file.append(".obj");
        if (fs::exists(fs::path(obj_file)))
        {
            target_objs.push_back(obj_file);
            tmp_list.push_back(obj_file);
        }
    }
}

/*********************************************************************/
/* Perform S2S Lowering to create host int.c and .cl files           */
/*********************************************************************/
void DoS2SLowering(Compiler& s2s,
                   std::vector<std::string>& input_c_files,
                   std::vector<std::string>& cl_chain,
                   std::vector<std::string>& h_int_list,
                   std::vector<std::string>& tmp_list)
{
    /* Create .int.c and .cl files */
    std::string cl_file;
    std::string int_file;
    for(auto infile : input_c_files)
    {
        /* Generate .int.c and .cl files */
        s2s.Compile(infile);
        cl_file = fs::path(infile).stem().string();
        int_file = cl_file;
        /* ASSUMPTION: At this stage the files that do not have 
         * target regions will not generate .cl files 
         * */
        /* Check to see whether .cl file has been generated 
         * and add it to clocl chain list 
         * */
        cl_file.append(".cl");
        if (fs::exists(fs::path(cl_file)))
        {
            cl_chain.push_back(cl_file);
            tmp_list.push_back(cl_file);
        }
        /* Check existence and add the generated int.c file name 
         * to host_int_files list 
         * */
        int_file.append(".int.c");
        if(fs::exists(fs::path(int_file)))
        {
            h_int_list.push_back(int_file);
            tmp_list.push_back(int_file);
        }
    }
}

/*********************************************************************/
/* Chain all generated .cl files together (if present) and           */
/* run CLOCL on the chained cl file with the target object files to  */
/* create a header file with a char array containing the cl program  */
/* binary. Encode the lenght of the char array and the array in a C  */
/* file and compile the c file to an object file for final stage link*/
/*********************************************************************/
void MakeCloclArtifacts(Compiler& clocl,
                        Compiler& h_link_cc,
                        std::vector<std::string>& cl_chain,
                        std::vector<std::string>& tmp_list)
{
    /* If no target regions exist in the code, no .cl files are generated  
     * In this case, generate a warning message and compile the 
     * int.c files for host execution only
     * */
    if (cl_chain.size() == 0)
    {
        std::cout << ">> WARNING: Either there were no target regions"
                  << " present in source files or there were errors in"
                  << " compiling target region source code."
                  << " Creating host only executable" 
                  << std::endl;
        full_host_execution = true;
    }else
    {   
        /* Chain together all .cl files into one .cl file for clocl to compile */
        std::string cl_filename_stem = 
                fs::path(chained_cl_file_name).stem().string();

        std::ofstream chained_cl_file_stream(chained_cl_file_name, 
                                             std::ios_base::binary
                                             );
        for(auto clfile : cl_chain)
        {
            std::ifstream cl_file_stream(clfile, std::ios_base::binary);
            chained_cl_file_stream << cl_file_stream.rdbuf();
            cl_file_stream.close();
        }
        chained_cl_file_stream.close();

        /* CLOCL: link DSP objs to create binary header file */
        clocl.set_link_lib_dependencies(target_objs);
        clocl.Link(chained_cl_file_name);
        tmp_list.push_back(chained_cl_file_name);

        /* CLOCL generates a .out file by default */
        /* Check if it exists and remove this if -k is not specified */
        std::string clocl_cl_out_file = cl_filename_stem + ".out";
        if (fs::exists(fs::path(clocl_cl_out_file)))
        {
            tmp_list.push_back(clocl_cl_out_file);
        }

        /* Check binary header file existence and add it to 
         * final link source files */
        std::string cl_bin_header_file = cl_filename_stem + ".dsp_h";
        std::string cl_bin_c_file = cl_filename_stem + ".c";
        std::string cl_bin_o_file = cl_filename_stem + ".o";
        if (fs::exists(fs::path(cl_bin_header_file)))
        {
            tmp_list.push_back(cl_bin_header_file);
            /* Create the C file */        
            fs::copy_file(fs::path(cl_bin_header_file),
                          fs::path(cl_bin_c_file), 
                          fs::copy_option::overwrite_if_exists
                          );

            if (fs::exists(fs::path(cl_bin_c_file)))
            {
                tmp_list.push_back(cl_bin_c_file);
                /* Append the debug variables to the C file */
                std::string runtime_dbg_str = runtime_dbg_var + " = ";
                std::string runtime_perf_str = runtime_perf_var + " = ";
                std::ofstream cl_bin_c_file_ostream (cl_bin_c_file, 
                                                     std::ios_base::app 
                                                     | std::ios_base::out
                                                     );
                /* Check if runtime debugging info is enabled and
                 * append hook for it */
                cl_bin_c_file_ostream << "bool "
                                      << runtime_dbg_str;
                if (runtime_debug)
                {
                    cl_bin_c_file_ostream << "true";
                }else
                {
                    cl_bin_c_file_ostream << "false";
                }
                cl_bin_c_file_ostream << ";\n";
                
                /* Check if runtime performance info is enabled and
                 * append hook for it */
                cl_bin_c_file_ostream << "bool "
                                      << runtime_perf_str;
                if (runtime_performance)
                {
                    cl_bin_c_file_ostream << "true";
                }else
                {
                    cl_bin_c_file_ostream << "false";
                }
                cl_bin_c_file_ostream << ";\n";
                
                /* Done appending. Close the file stream */
                cl_bin_c_file_ostream.close();

                /* Compile C file with host link cc compiler */
                h_link_cc.Compile(cl_bin_c_file);
                if (fs::exists(fs::path(cl_bin_o_file)))
                {
                    /* Add compiled .o file to final link objs */
                    host_objs.push_back(cl_bin_o_file);         
                    tmp_list.push_back(cl_bin_o_file);
                }else
                {
                    std::cerr << ">> ERROR: " << cl_bin_o_file
                              << " not generated " 
                              << std::endl;
                    exit(EXIT_FAILURE);
                }
            }else
            {
                std::cerr << ">> ERROR: " << cl_bin_c_file 
                          << " not generated " 
                          << std::endl;
                exit(EXIT_FAILURE);
            }
        }else
        {
            std::cerr << ">> ERROR: " << cl_bin_header_file 
                      << " not generated " 
                      << std::endl;
            exit(EXIT_FAILURE);
        }
    }
}

/*********************************************************************/
/* Perform final linking into fat binary                             */
/* If no cl files were generated then full host execution is enabled */
/* In this case generate ARM only executable                         */
/*********************************************************************/
void DoFinalLink(Compiler& h_cc,
                 Compiler& h_link_cc,
                 std::vector<std::string>& input_c_files,
                 std::vector<std::string>& h_int_files,
                 std::vector<std::string>& tmp_list)
{
    /* Link all .int.o files with TIOpenMPACC lib and OpenCL lib with DSP
     * binary header to create fat binary */
    /* If no target regions present, i.e. no .cl files have been generated, 
     * create host binary only or else create fat binary with target code */
    if (full_host_execution)
    {
        /* Compile input files */
        std::string in_o_file;
        for (auto infile : input_c_files)
        {
            h_cc.Compile(infile);
            in_o_file = fs::path(infile).stem().string() + ".o";
            if (fs::exists(fs::path(in_o_file)))
            {
                tmp_list.push_back(in_o_file);
                user_host_objs.push_back(in_o_file);
            }
        }
        h_link_cc.Link(user_host_objs, exe_name); 
    }
    else
    { 
        /* Compile lowered int.c files */
        std::string int_o_file;
        for (auto int_file : h_int_files)
        {
            h_cc.Compile(int_file);
            int_o_file = fs::path(int_file).stem().string();
            int_o_file.append(".o");
            if(fs::exists(fs::path(int_o_file)))
            {
                host_objs.push_back(int_o_file);
                tmp_list.push_back(int_o_file);
            }
        }

        /* If no host objs were produced, then there must 
         * have been errors in compiling int.c files */
        if (user_host_objs.size() == host_objs.size())
        {
            std::cerr << ">> ERROR: Compilation of host objects failed" 
                      << std::endl;
            exit(EXIT_FAILURE);
        }

        h_link_cc.Link(host_objs, exe_name);
    }
}

/*********************************************************************/
/* Unless -k is set, remove all intermediate files and if debug is   */
/* set then report if the fat binary was created or not              */
/*********************************************************************/
void CleanUp(std::vector<std::string>& tmp_list)
{
    /* Remove intermediate files unless -v specified */
    if(!keep_tmp_files)
    {
        for (auto tmp_file : tmp_list)
        {
            fs::remove(fs::path(tmp_file));
        }
    }

    if (debug)
    {
        if(fs::exists(fs::path(exe_name)))
            std::cout << ">> DEBUG: Fat binary " << exe_name 
                      << " created successfully." 
                      << std::endl;
        else 
            std::cerr << ">> ERROR: Fat binary not created." 
                      << std::endl;
    }
}
