/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file elfutils.cpp
 * \brief Wrapper for libelf utilities to examine ELF object files 
 */

#include "elfutils.h" 

/*********************************************************************/
/* Initialize libelf                                                 */
/*********************************************************************/
void ElfInit()
{
    if (elf_version(EV_CURRENT) == EV_NONE)
        std::cerr << ">> ERROR: libelf init failed" << std::endl;
}

/*********************************************************************/
/* Check if given file is an elf object file or an elf archive       */
/*********************************************************************/
bool IsElf(const std::string& file)
{
    int fd;
    Elf *elf_file;
   
    if (( fd = open(file.c_str(), O_RDONLY , 0)) < 0)
        std::cerr << ">> ERROR: " << file << " open failed" << std::endl;

    if (( elf_file = elf_begin(fd , ELF_C_READ, NULL)) == NULL )
        std::cerr << ">> ERROR: libelf: elf begin failed" << std::endl; 

    if (elf_kind(elf_file) != ELF_K_ELF && elf_kind(elf_file) != ELF_K_AR)
        return false;
    
    elf_end(elf_file);
    close(fd);

    return true;
}

/*********************************************************************/
/* Get the machine id of the given elf file                          */
/* If file is an archive, then get machine id of first archive member*/
/*********************************************************************/
int GetElfMachineID(const std::string& file)
{
    int fd;
    Elf *elf_file, *elf_ar_member;
    Elf_Arhdr *elf_ar_member_hdr;
    GElf_Ehdr elf_hdr;
    int machine_id = EM_ARM;
    bool got_machine_id = false;

    if (( fd = open(file.c_str(), O_RDONLY , 0)) < 0)
    {
        std::cerr << ">> ERROR: " << file << " open failed" << std::endl;
        exit(EXIT_FAILURE);
    }
    if (( elf_file = elf_begin(fd , ELF_C_READ, NULL)) == NULL )
    {
        std::cerr << ">> ERROR: libelf: elf begin failed" << std::endl; 
        exit(EXIT_FAILURE);
    }
    if (elf_kind(elf_file) == ELF_K_ELF)
    {
        /* If it is an elf object file, then get its header */
        if (gelf_getehdr(elf_file, &elf_hdr) == NULL)
        {
            std::cerr << ">> ERROR: libelf: gelf getehdr failed" << std::endl; 
            exit(EXIT_FAILURE);
        }
        machine_id = elf_hdr.e_machine; 
    
    }else if (elf_kind(elf_file) == ELF_K_AR)
    { 
    
        while (!got_machine_id)
        {
            /* If it is an archive, then inspect an archive member */
            if (( elf_ar_member = elf_begin(fd , ELF_C_READ, elf_file)) == NULL )
            {
                std::cerr << ">> ERROR: libelf: elf begin failed" << std::endl;
                exit(EXIT_FAILURE);
            }
            /* Get elf archive header of the archive member */
            if ((elf_ar_member_hdr = elf_getarhdr(elf_ar_member)) == NULL)
            {
                std::cerr << ">> ERROR: libelf: elf getarhdr failed" << std::endl;
                exit(EXIT_FAILURE);
            }
            
            /* Check if archive member is not empty or is not a symbol table 
             * If it is, go to process the next member */
            std::string elf_ar_name = std::string(elf_ar_member_hdr->ar_name);
            if (elf_ar_name == "/"
                || elf_ar_name == "//"
                || elf_ar_name == "__.SYMDEF")
            {
                elf_next(elf_ar_member);
                elf_end(elf_ar_member);
                continue;
            }

            /* Get elf header of the archive member */
            if (gelf_getehdr(elf_ar_member, &elf_hdr) == NULL)
            {
                std::cerr << ">> ERROR: libelf: gelf getehdr failed" << std::endl; 
                exit(EXIT_FAILURE);
            }
            /* Get machine ID of first archive member 
             * ASSUMPTION: This is not a mixed archive and all members have 
             *             the same machine ID */
            machine_id = elf_hdr.e_machine; 
            elf_end(elf_ar_member);
            got_machine_id = true;
        }
    }

    elf_end(elf_file);
    close(fd);

    return machine_id;
}
