/******************************************************************************
 * Copyright (c) 2013-2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/
#include <iostream>
#include <cstdlib>
#include <iomanip>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <omp.h>
#include "ti_omp_device.h"

using namespace std;

/* ======================================================================== */
/*  Global Variables                                                        */
/* ======================================================================== */
float *A; 
float *B;  
float *C_host; 
float *C_device; 
float *Golden;

#define RAND_MAX_A                  5
#define RAND_MAX_B                  5
#define EPISILON                    0.00001
int M                               = 10;
int N                               = 10;
int P                               = 10;
int cout_text_width                 = 55;
int cout_precision                  = 8;
bool print_matrices                 = false;
bool show_timing                    = false;
bool show_array_sizes               = false;
bool host_omp_matmul                = false;
bool host_omp_matmul_error_check    = false;

/* ======================================================================== */
/*  Function Headers                                                        */
/* ======================================================================== */
void PrintUsageAndExit();
void PrintMatrix(float *mat, int rows, int cols);
void HandleOptions(int argc, char* argv[]);
void ShowArraySizes();
void InitArrays();
double MatmulNaiveHost();
double MatmulNaiveHostOMP();
void CheckForErrors(float* Mat);
void PrintBlock(string text);

extern "C" {
    extern void MatmulNaiveTarget(float* A, 
                                  float* B, 
                                  float* C,
                                  int M, 
                                  int N, 
                                  int P
                                  );
}


/* ======================================================================== */
/*  MAIN                                                                    */
/* ======================================================================== */
int main(int argc, char* argv[])
{
    /* ---------------------------------------------------------------- */
    /*  Handle command line options to get M,N,P                        */
    /* ---------------------------------------------------------------- */
    HandleOptions(argc,argv);

    if (show_array_sizes)
        ShowArraySizes(); 

    double start = 0, end =0;
    double matmul_naive_host_time = 0;
    double matmul_naive_host_omp_time = 0;
    double matmul_naive_device_time = 0;
    int acc_device_id = 0;
    /* ---------------------------------------------------------------- */
    /*  Allocate arrays:                                                */
    /*  Shared CMEM: A, B, C_device                                     */
    /*  Host Memory: C_host, Golden                                     */
    /* ---------------------------------------------------------------- */
    A           = (float*) __TI_omp_device_alloc(acc_device_id,  
                                                 TI_OMP_MEM_DDR, 
                                                 M*N*sizeof(float)
                                                 );
    B           = (float*) __TI_omp_device_alloc(acc_device_id, 
                                                 TI_OMP_MEM_DDR, 
                                                 N*P*sizeof(float)
                                                 );
    C_device    = (float*) __TI_omp_device_alloc(acc_device_id, 
                                                 TI_OMP_MEM_DDR, 
                                                 M*P*sizeof(float)
                                                 );
    C_host      = new float[M*P];
    Golden      = new float[M*P];

    /* ---------------------------------------------------------------- */
    /*  Initialized input arrays with random test data.                 */
    /* ---------------------------------------------------------------- */
    InitArrays();

    /* ---------------------------------------------------------------- */
    /* Host: Do A*B = Golden for verification                           */
    /* ---------------------------------------------------------------- */
    matmul_naive_host_time = MatmulNaiveHost();

    /* ---------------------------------------------------------------- */
    /* Host: Do A*B = C using OpenMP                                    */
    /* ---------------------------------------------------------------- */
    if (host_omp_matmul)
        matmul_naive_host_omp_time = MatmulNaiveHostOMP();
    
    /* ---------------------------------------------------------------- */
    /* Perform error checking on host OpenMP matmul result              */
    /* ---------------------------------------------------------------- */
    if (host_omp_matmul && host_omp_matmul_error_check)
        CheckForErrors(C_host);

    /* ---------------------------------------------------------------- */
    /* Device: Do A*B = C using OpenMP Accelerator Model                */
    /* ---------------------------------------------------------------- */
    start = omp_get_wtime();
    MatmulNaiveTarget(A, B, C_device, M, N, P);
    end = omp_get_wtime();
    if (print_matrices)
    {
        cout << "C_device:" << endl;
        PrintMatrix(C_device, M, P);
    }
    
    matmul_naive_device_time = end - start;

    if (show_timing)
    {
        cout << setw(cout_text_width)
             << "Device naive matmul (8 DSP Cores, OpenMPAcc)" 
             << "= "
             << setprecision(cout_precision)
             << matmul_naive_device_time << endl;
        
        PrintBlock("SPEED-UP");

        if (host_omp_matmul)
        {
             double speedup_host_omp_v_host = matmul_naive_host_time 
                                              / matmul_naive_host_omp_time;
             cout << setw(cout_text_width)
                  << "Host OpenMP (vs. Matmul Naive Host)"
                  << "= "
                  << setprecision(cout_precision)
                  << speedup_host_omp_v_host
                  << endl;

             double speedup_host_omp_v_device = matmul_naive_host_omp_time 
                                                / matmul_naive_device_time;
             cout << setw(cout_text_width)
                  << "Device OpenMPAcc (vs. Matmul Naive Host OpenMP)"
                  << "= "
                  << setprecision(cout_precision)
                  << speedup_host_omp_v_device
                  << endl;
        }
        double speedup_host_v_device = matmul_naive_host_time 
                                       / matmul_naive_device_time;
        cout << setw(cout_text_width)
             << "Device OpenMPAcc (vs. Matmul Naive Host)"
             << "= " 
             << setprecision(cout_precision)
             << speedup_host_v_device
             << endl;
    }

    /* ---------------------------------------------------------------- */
    /* Perform error checking on Device OpenMPAcc matmul result         */
    /* ---------------------------------------------------------------- */
    CheckForErrors(C_device);

    /* ---------------------------------------------------------------- */
    /*  Free Arrays                                                     */
    /* ---------------------------------------------------------------- */
    __TI_omp_device_free(acc_device_id, A);
    __TI_omp_device_free(acc_device_id, B);
    __TI_omp_device_free(acc_device_id, C_device);
    delete [] C_host;
    delete [] Golden;

    return 0;
}

/**
 * \brief Print a pretty block with the 'text' enclosed
 * \param text String to enclose in the pretty block
 */
void PrintBlock(string text)
{
    text = "| " + text;
    cout << left
         << setw(cout_text_width) << setfill('-')
         << "" << endl
         << setfill(' ')
         << text 
         << right << setw(cout_text_width - text.length()) << setfill(' ') 
         << "|" << endl
         << left << setw(cout_text_width) << setfill('-')
         << "" << endl
         << setfill(' ');
}

/**
 * \brief Print matrix sizes of A,B,C in bytes, Kbytes, Mbytes
 */
void ShowArraySizes()
{
    PrintBlock("MATRIX SIZES");
    cout << "M,N,P : " << M << "," << N << "," << P << endl;
    float size_A = M*N*sizeof(float)*1.0;
    float size_B = N*P*sizeof(float)*1.0;
    float size_C = M*P*sizeof(float)*1.0;
    cout << "Size of A[M,N]: " << size_A << " b | " 
                               << size_A/1024.0 << " Kb | " 
                               << size_A/1024.0/1024.0 << " Mb" 
                               << endl;
    cout << "Size of B[N,P]: " << size_B << " b | " 
                               << size_B/1024.0 << " Kb | " 
                               << size_C/1024.0/1024.0 << " Mb" 
                               << endl;
    cout << "Size of C[M,P]: " << size_C << " b | " 
                               << size_C/1024.0 << " Kb | " 
                               << size_C/1024.0/1024.0 << " Mb" 
                               << endl;
}

/**
 * \brief Initialize matrix A,B with random numbers 
 */
void InitArrays()
{
    double start = 0, end = 0;
    int i, j;
    
    start = omp_get_wtime();
    #pragma omp parallel shared(A,B) private(i,j) firstprivate(M,N,P)
    {
        /* Initialize Random Number Seed */
        srand(time(NULL));
        #pragma omp for collapse(2)
        for (i=0; i < M; i++)
            for (j=0; j < N; j++)
                A[i*N + j] = (rand() % RAND_MAX_A + 1) * 1.0; 
        
        #pragma omp for collapse(2)
        for (i=0; i < N; i++)
            for (j=0; j < P; j++)
                B[i*P + j] = (rand() % RAND_MAX_B + 1) * 1.0;
    }
    end = omp_get_wtime();

    if (print_matrices)    
    {
        cout << "A:" << endl;
        PrintMatrix(A, M, N);

        cout << "B:" << endl;
        PrintMatrix(B, N, P);
    }

    if (show_timing)
    {
        PrintBlock("COMPUTATION TIME (in Seconds)");
        cout << setw(cout_text_width) << left
             << "Init A,B with random numbers (4 ARM Cores, OpenMP)" 
             << "= "
             << setprecision(cout_precision)
             << end - start << endl; 
    }
}

/**
 * \brief Perform Golden = A*B on a single Host (ARM) core
 *        Golden is later used an a verification/error checking matrix
 */
double MatmulNaiveHost()
{
    double start = 0, end = 0, elapsed = 0;
    start = omp_get_wtime();
    for (int i = 0; i < M; ++i) 
        for (int j = 0; j < P; ++j)
        {
            float sum = 0;
            for (int k = 0; k < N; ++k)
            {
                sum += A[i*N +k] * B[k*P + j];
            }
            Golden[i*P + j] = sum;
        }
    end = omp_get_wtime();

    if (print_matrices)
    {
        cout << "Golden:" << endl;
        PrintMatrix(Golden, M, P);
    }
    
    elapsed = end - start;
    if (show_timing)
    {
        cout << setw(cout_text_width) << left
             << "Host naive matmul (1 ARM Core)" 
             << "= "
             << setprecision(cout_precision)
             << elapsed << endl; 
    }

    return elapsed;
}

/**
 * \brief Perform C_host = A*B across 4 Host (ARM) cores
 */
double MatmulNaiveHostOMP()
{
    double start = 0, end = 0, elapsed = 0;
    int i, j, k;
    float sum;
    start = omp_get_wtime();
    #pragma omp parallel for collapse(2) private(i,j,k,sum) firstprivate(M,N,P) shared(A,B,C_host)
    for (i = 0; i < M; ++i) 
        for (j = 0; j < P; ++j)
        {
            sum = 0;
            for (k = 0; k < N; ++k)
            {
                sum += A[i*N +k] * B[k*P + j];
            }
            C_host[i*P + j] = sum;
        }
    end = omp_get_wtime();
    
    if (print_matrices)
    {
        cout << "C_host:" << endl;
        PrintMatrix(C_host, M, P);
    }
    
    elapsed = end - start;
    if (show_timing)
    {
        cout << setw(cout_text_width) << left
             << "Host naive matmul (4 ARM Cores, OpenMP)"
             << "= "
             << setprecision(cout_precision)
             << elapsed << endl; 
    }
    
    return elapsed;
}

/**
 * \brief Check given matrix for errors against Golden verification matrix
 *        Use all Host (ARM) cores to do error checking
 * \param Mat Matrix to check errors in
 *        Mat must have dimensions M*P
 */
void CheckForErrors(float* Mat)
{   
    int num_errors = 0, i, j;
    const int print_nerrors = 12;
    
    #pragma omp parallel for collapse(2) shared(Golden, Mat) private(i,j) firstprivate(M,P)
    for (i=0; i<M; i++)
        for (j=0; j<P; j++)
        {
            if (Golden[i*P + j] - Mat[i*P + j] < -EPISILON 
                || Golden[i*P + j] - Mat[i*P + j] > EPISILON) 
            { 
                if((num_errors += 1) < print_nerrors)
                    printf("Error [%d,%d]: %f <==> %f\n", 
                           i, 
                           j, 
                           Golden[i*P + j], 
                           Mat[i*P + j]
                           );
            }
        }

    if (num_errors > 0) 
        cout << "FAIL with " << num_errors << " errors!\n";
    else                 
        cout << "PASS!" << endl; 
}

/**
 * \brief Print Help Message and Exit
 */
void PrintUsageAndExit()
{
    cout << "Matrix A[M,N] x B [N,P] = C [M,P]" << endl
         << "Default value of M,N,P is " << M << endl 
         << "Usage: matmul_naive [options] " << endl
         << "Options: " << endl
         << "-M arg : Number of rows for array A" << endl
         << "-N arg : Number of cols for array A, rows for array B" << endl
         << "-P arg : Number of cols for array B" << endl
         << "-a     : Peform host OpenMP naive matmul" << endl
         << "-e     : Peform host OpenMP naive matmul error checking" << endl
         << "-d     : Enable debug printing of matrices A, B, Golden, C" << endl
         << "-t     : Show timing and speed-up information" << endl
         << "-s     : Show array size information" << endl
         << "-h     : Show help message" 
         << endl;
    exit(0);
}

/**
 * \brief Print the given matrix
 * \param mat   Pointer to matrix
 * \param rows  Number of rows in matrix
 * \param cols  Number of cols in matrix
 */
void PrintMatrix(float *mat, int rows, int cols)
{
    for (int i=0; i<rows; i++)
    {
        for (int j=0; j<cols; j++)
        {
            cout << setprecision(cout_precision) 
                 << setw(cout_precision+1) 
                 << mat[i*cols + j] 
                 << " ";
        }
        cout << endl;
    }
}

/**
 * \brief Handle command line options
 * \param argc  Number of options
 * \param argv  Array containing option strings
 */
void HandleOptions(int argc, char* argv[])
{
    int c;

    if (argc == 1)
        return;
    
    while ((c = getopt (argc, argv, "M:N:P:dthsae")) != -1)
    {
        switch(c)
        {
            case 'M':
                M = atoi(optarg);
                break;
            case 'N':
                N = atoi(optarg);
                break;
            case 'P':
                P = atoi(optarg);
                break;
            case 'd':
                print_matrices = true;
                break;
            case 't':
                show_timing = true;
                break;
            case 's':
                show_array_sizes = true;
                break;
            case 'a':
                host_omp_matmul = true;
                break;
            case 'e':
                host_omp_matmul_error_check = true;
                break;
            case 'h':
                PrintUsageAndExit();
                break;
            default:
                PrintUsageAndExit();
        }
    }
}
