/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file accelerator.cc
 * \brief OpenCL Accelerator Wrapper  
 */

#include "accelerator.h"

/*********************************************************************/
/* OpenCL Program binary char array and len generated by CLOCL       */
/* These will be linked into the final executable by CLACC           */
/* These symbols are declared weak so linkage does not fail for host */
/* only execution mode when __TI_CLACC_KERNEL.c is not generated     */
/*********************************************************************/
extern unsigned int __attribute__((weak)) __TI_CLACC_KERNEL_dsp_bin_len;
extern char __attribute__((weak)) __TI_CLACC_KERNEL_dsp_bin[];

/*********************************************************************/
/* Weak hooks to enable debugging and performance information        */
/*********************************************************************/
extern bool __attribute__((weak)) __TI_show_debug_;
extern bool __attribute__((weak)) __TI_show_perf_;

/*********************************************************************/
/* Constructor: Do initial setup                                     */
/*********************************************************************/
Accelerator::Accelerator(int type)
{
    init_ = false;
    context_ = cl::Context(type);
    devices_ = context_.getInfo<CL_CONTEXT_DEVICES>();
    timer_resolution_ = 0;
    /* Check if the addresses of the variables are null, if not then
     * they are defined and then use their values */
    if (&__TI_show_debug_)
        debug_ = __TI_show_debug_;
    else
        debug_ = false;

    if(&__TI_show_perf_)
        perf_ = __TI_show_perf_;
    else
        perf_ = false;
}

/*********************************************************************/
/* Initialize global generic OpenCL variables                        */
/*********************************************************************/
void Accelerator::InitializeCL(int device_id)
{
    if(!init_)
    {   
        try
        {
            device_id_ = device_id;
            queue_ = cl::CommandQueue(context_, 
                                      devices_[device_id], 
                                      CL_QUEUE_PROFILING_ENABLE
                                      );
            devices_[device_id].getInfo(CL_DEVICE_PROFILING_TIMER_RESOLUTION, 
                                        &timer_resolution_
                                        );
            devices_[device_id].getInfo(CL_DEVICE_MEM_BASE_ADDR_ALIGN,
                                        &base_addr_align_
                                        );
            if (debug_)
            {
                std::string str;
                devices_[device_id].getInfo(CL_DEVICE_NAME, &str);
                std::cout << ">> TI_DEBUG: Accelerator Device: " 
                          << str 
                          << std::endl;
                std::cout << ">> TI_DEBUG: Device Memory Base Address Alignment: " 
                          << base_addr_align_
                          << std::endl;
            }

            /******************************************************************/
            /* Extract program binary from embedded char array                */ 
            /* and build program if the weak symbols are defined              */ 
            /******************************************************************/
            if (&__TI_CLACC_KERNEL_dsp_bin_len && __TI_CLACC_KERNEL_dsp_bin)
                BuildProgramFromBinary(__TI_CLACC_KERNEL_dsp_bin,
                                       __TI_CLACC_KERNEL_dsp_bin_len
                                      );

            init_ = true;
        }catch (Error err) 
        { 
            std::cerr << ">> ERROR: Accelerator initialization: " 
                      << err.what() << "(" << err.err() << ")" 
                      << std::endl; 
        }
        
        if (debug_)
        {
            std::cout << ">> TI_DEBUG: Iniitialized Accelerator Device" 
                      << std::endl;
        }
 
    }

}

/*********************************************************************/
/* Make a target update for the given list items                     */
/*********************************************************************/
void Accelerator::MakeTargetUpdate( int mapnum,
                                    void **hostaddrs,
                                    unsigned int *sizes,
                                    unsigned char *kinds
                                    )
{
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET_UPDATE: Number of list items: " 
                  << mapnum 
                  << std::endl;
    }
    bool update_performed = false;
    
    /* Extract target update arg pointers */
    std::vector<void*> tupdate_arg_ptrs;
    ExtractArgPointers(hostaddrs, mapnum, tupdate_arg_ptrs);
   
    DataSlice arg_slice;
    for (int i=0; i< mapnum; i++)
    { 
        /* Create DataSlice */
        arg_slice = DataSlice(tupdate_arg_ptrs[i],
                              sizes[i],
                              static_cast<MapKind>(kinds[i]),
                              debug_,
                              perf_
                              );

        update_performed = UpdateTdataSlices(arg_slice);
        
        if (update_performed)       
            continue;
        
        /*-------------------------------------------------------------------*/
        /* At this point, arg_slice is guaranteed to not be present in       */
        /* the tdata slices, so check for it in tmalloc slices               */
        /*-------------------------------------------------------------------*/
       
        update_performed = UpdateTmallocSlices(arg_slice);
    }
    
    if (!update_performed)
    {
        /*-------------------------------------------------------------------*/
        /* At this point, the arg_slice is guaranteed to not partially       */ 
        /* overlap any unique slice and is not contained within any of the   */
        /* unique slices i.e. it is a unique slice itself i.e. cannot be     */
        /* updated as it wasn't declared before                              */
        /*-------------------------------------------------------------------*/
        std::cerr << "WARNING: TARGET_UPDATE: Data section not present in "
                  << "TARGET_DATA sections or have not been TI_OMP_DEVICE_ALLOC'd:"
                  << std::endl << "\t"
                  << arg_slice
                  << std::endl;
    }
    
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET_UPDATE: Done" 
                  << std::endl;
    }
}

/*********************************************************************/
/* Update slice if found in tdata slices                             */
/*********************************************************************/
bool Accelerator::UpdateTdataSlices(DataSlice& arg_slice)
{
    /* The direction of the update i.e. either MAP_TO or MAP_FROM */ 
    MapKind update_motion = arg_slice.kind();
    DataSliceID arg_slice_id = arg_slice.id();

    /* Check if DataSlice is already present in tdata_unique_slices */
    for (auto& kv : tdata_unique_slices_)
    {
        DataSlice& slice = kv.second;
        /* If start, size is the same, then use the same buffer 
         * NOTE: The child kind is disregarded
         * REF: OMP 4.0 Spec: Page 177: Lines 17-21 */
        if (slice == arg_slice)
        {
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET_UPDATE: "
                          << "Handling section: " 
                          << std::endl << "\t"
                          << slice 
                          << std::endl;
            }
            slice.Update(queue_, update_motion); 
            return true;
        }   
        else if (slice.Contains(arg_slice))
        {
            /* Test whether arg_slice is present in tdata_sub_slices_ */
            if(tdata_sub_slices_.count(arg_slice_id) > 0)
            {
                if(debug_)
                {
                    std::cout << ">> TI_DEBUG: TARGET_UPDATE: "
                              << "Handling section: " 
                              << std::endl << "\t"
                              << tdata_sub_slices_[arg_slice_id] 
                              << std::endl;
                }
                tdata_sub_slices_[arg_slice_id].Update(queue_,
                                                       update_motion
                                                       );
                return true;
            }

            /* arg_slice is *fully* contained in a unique slice,
             * make it a sub slice of the unique slice and update it */ 
            arg_slice.MakeSubSliceOf(slice);
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET_UPDATE: Handling section: " 
                          << std::endl << "\t"
                          << arg_slice 
                          << std::endl;
            }
            arg_slice.Update(queue_, update_motion); 
            return true;
        }
    }

    return false;
}

/*********************************************************************/
/* Update slice if found in tmalloc slices                           */
/*********************************************************************/
bool Accelerator::UpdateTmallocSlices(DataSlice& arg_slice)
{
    /* The direction of the update i.e. either MAP_TO or MAP_FROM */ 
    MapKind update_motion = arg_slice.kind();

    /* Check if DataSlice is already present in tmalloc_unique_slices */
    for (auto& kv : tmalloc_unique_slices_)
    {
        DataSlice& slice = kv.second;
        /* If start, size is the same, then use the same buffer 
         * NOTE: The child kind is disregarded
         * REF: OMP 4.0 Spec: Page 177: Lines 17-21 */
        if (slice == arg_slice)
        {
            if(debug_)
            {
                std::cout << "TI_DEBUG: TARGET_UPDATE: Handling section: " 
                          << std::endl << "\t"
                          << slice 
                          << std::endl;
            }
            /* Make update on cmem slice */
            if (update_motion == MAP_TO)
                CMEM_cacheWb(slice.start(), slice.size());
            else if (update_motion == MAP_FROM)
                CMEM_cacheInv(slice.start(), slice.size());

            return true;
        }
        else if (slice.Contains(arg_slice))
        {
            /* Check if arg_slice is *fully* contained in 
             * a unique slice. If so, make the update */ 
            if(debug_)
            {
                std::cout << "TI_DEBUG: TARGET_UPDATE: Handling section: " 
                          << std::endl << "\t"
                          << arg_slice 
                          << std::endl;
            }
            /* NOTE: No need to check in tmalloc_sub_slices_ here
             * as only the start address and size are required to 
             * make the cacheInv */
            
            /* Make update on cmem slice */
            if (update_motion == MAP_TO)
                CMEM_cacheWb(arg_slice.start(), arg_slice.size());
            else if (update_motion == MAP_FROM)
                CMEM_cacheInv(arg_slice.start(), arg_slice.size());

            return true;  
        }
    }

    return false;
}

/*********************************************************************/
/* Setup data items specific to a target data region                 */
/*********************************************************************/
void Accelerator::SetupTargetDataRegion( int mapnum,
                                         void **hostaddrs,
                                         unsigned int *sizes,
                                         unsigned char *kinds
                                        )
{
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET_DATA: Initializing.."
                  << std::endl
                  << ">> TI_DEBUG: TARGET_DATA: Number of list items: " 
                  << mapnum 
                  << std::endl;
    }
    
    bool slice_found;
    /* Extract target data arg pointers */
    std::vector<void*> tdata_arg_ptrs;
    ExtractArgPointers(hostaddrs, mapnum, tdata_arg_ptrs);
    
    for (int i=0; i< mapnum; i++)
    {
        slice_found = false;    
        /* Create DataSlice */
        DataSlice arg_slice = DataSlice(tdata_arg_ptrs[i],
                                        sizes[i],
                                        static_cast<MapKind>(kinds[i]),
                                        debug_,
                                        perf_
                                        );
        
        DataSliceID arg_slice_id = arg_slice.id();
        MapKind arg_slice_kind = arg_slice.kind(); 
        
        /* Check for arg_slice presence in tdata slices that have already 
         * been created by parent TARGET_DATA call. If arg_slice is 
         * contained in a parent TARGET_DATA section, then it is made into
         * a sub-slice of the parent slice and its device_buffer is 
         * initialized to be a subBuffer of the parent device_buffer. It is
         * then added to the sub_slice list */
        slice_found = CheckForSlice(arg_slice, 
                                    tdata_unique_slices_,
                                    tdata_sub_slices_,
                                    "TARGET_DATA"
                                    );       
        /* If arg_slice was found to already be declared in a parent 
         * TARGET_DATA call or be contained within a parent TARGET_DATA 
         * section, then no need to check for it in other lists as it has 
         * already been handled 
         * Refer to OpenMP 4.0 Spec Page 179, Lines 6-8 */
        if(slice_found)
            continue;
        
        /*-------------------------------------------------------------------*/
        /* At this point, arg_slice is guaranteed to not be present in       */
        /* the tdata slices, so check in tmalloc slices                      */
        /*-------------------------------------------------------------------*/
        slice_found = CheckForSlice(arg_slice,
                                    tmalloc_unique_slices_,
                                    tmalloc_sub_slices_,
                                    "TARGET_DATA"
                                    );

        /* Same logic as above, but for TI_OMP_DEVICE_ALLOC'd sections 
         * Of course, this isn't officially part of the OpenMP 4.0 spec */
        if(slice_found)
            continue;
        
        /*-------------------------------------------------------------------*/
        /* At this point, arg_slice is guaranteed to not be present in       */
        /* tdata and tmalloc slices                                          */
        /*-------------------------------------------------------------------*/
       
        CheckSliceOverlap(arg_slice, "TARGET_DATA");
        
        /*-------------------------------------------------------------------*/
        /* At this point, the arg_slice is guaranteed to not partially       */ 
        /* overlap any unique slice and is not contained within any of the   */
        /* unique slices i.e. it is a unique slice itself                    */
        /*-------------------------------------------------------------------*/
        if(debug_)
        {
            std::cout << ">> TI_DEBUG: TARGET_DATA: Creating data section: " 
                      << std::endl << "\t"
                      << arg_slice 
                      << std::endl;
        }
        /* Initialize its CL buffer */
        arg_slice.InitDeviceBuffer(context_);
        /* If MAP_TO kind then start a non-blocking write to device buffer */
        if (arg_slice_kind == MAP_TO 
            || arg_slice_kind == MAP_TO_FROM)
        {
            arg_slice.WriteToDevice(queue_, false);
        }
        /* Add it to the unique slice list */
        tdata_unique_slices_[arg_slice_id] = arg_slice;
    }
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET_DATA: Initialization complete" 
                  << std::endl;
    }
}

/*********************************************************************/
/* Check presence of slice in the given unique and sub lists         */
/* and handle if found                                               */
/*********************************************************************/
bool Accelerator::CheckForSlice(DataSlice& arg_slice,
                                std::unordered_map<DataSliceID,DataSlice>& unique_slices,
                                std::unordered_map<DataSliceID,DataSlice>& sub_slices,
                                const std::string& region)
{
    DataSliceID arg_slice_id = arg_slice.id();

    /* Check if DataSlice is already present in unique_slices */
    for (auto& kv : unique_slices)
    {
        DataSlice& slice = kv.second;
        /* If start, size is the same, then use the same buffer 
         * NOTE: The child kind is disregarded
         * REF: OMP 4.0 Spec: Page 177: Lines 17-21 */
        if (slice == arg_slice)
        {
            /* NOTE: No data transfer or CMEM_cacheWb made as data 
             * already exists on device */
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET_DATA: Found matching "
                          << "parent " << region << " section: " 
                          << std::endl << "\t"
                          << slice 
                          << std::endl;
            }
            return true;
        }
        else if (slice.Contains(arg_slice))
        {
            /* Test whether arg_slice is present in sub_slices */
            if(sub_slices.count(arg_slice_id) > 0)
            {
                /* NOTE: No data transfer made as data already exists 
                 * on device. In case of CMEM data, it will be written
                 * back upon entry to target region */
                if(debug_)
                {
                    std::cout << ">> TI_DEBUG: TARGET_DATA: "
                              << "Found matching " 
                              << region << " sub-section: "
                              << std::endl << "\t"
                              << sub_slices[arg_slice_id] 
                              << std::endl;
                }
                return true;
            }
            /* arg_slice is *fully* contained in a unique slice,
             * make it a sub slice of the unique slice */ 
            arg_slice.MakeSubSliceOf(slice);

            /* NOTE: No data transfer made as data already exists 
             * on device. In case of CMEM data, it will be written back 
             * upon entry to target region */

            /* Copy it to list of sub slices */
            sub_slices[arg_slice_id] = arg_slice;

            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET_DATA: "
                          << "Created sub-section" 
                          << std::endl;
            }
            return true;
        }
    }

    return false;
}

/*********************************************************************/
/* Check for overlap with tdata and tmalloc slices and exit if found */
/*********************************************************************/
void Accelerator::CheckSliceOverlap(DataSlice& arg_slice, std::string region)
{
    /* Check if any part of the slice is partially overlapping with a 
     * slice in the tdata or tmalloc unique lists i.e. check whether the 
     * slice start or slice end is present in any of the unique 
     * list slices */
    for (auto& kv : tdata_unique_slices_)
    {
        DataSlice& slice = kv.second;
        if (slice.Contains(arg_slice.start())
                || slice.Contains(arg_slice.end()))
        {
            std::cerr << ">> ERROR: " << region << ": Target data slice " 
                      << arg_slice << std::endl 
                      << "\t  has overlapping segment with parent "
                      << "target data section " 
                      << std::endl << "\t"
                      << slice 
                      << std::endl;
            exit(EXIT_FAILURE);
        }
    }

    for (auto& kv : tmalloc_unique_slices_)
    {
        DataSlice& slice = kv.second;
        if (slice.Contains(arg_slice.start())
                || slice.Contains(arg_slice.end()))
        {
            std::cerr << ">> ERROR: " << region << ": Target malloc slice " 
                      << arg_slice << std::endl 
                      << "\t  has overlapping segment with "
                      << "target malloc section " 
                      << std::endl << "\t"
                      << slice 
                      << std::endl;
            exit(EXIT_FAILURE);
        }
    }
}

/*********************************************************************/
/* Destroy data for a target data region                             */
/*********************************************************************/
void Accelerator::DestroyTargetDataRegion( int mapnum,
                                           void** hostaddrs,
                                           unsigned int* sizes,
                                           unsigned char* kinds
                                         )
{
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET_DATA_END: Closing region.." 
                  << std::endl
                  << ">> TI_DEBUG: TARGET_DATA_END: Number of list items: "
                  << mapnum
                  << std::endl;
    }

    /* Extract target data arg pointers */
    std::vector<void*> tdata_arg_ptrs;
    ExtractArgPointers(hostaddrs, mapnum, tdata_arg_ptrs);

    for (int i=0; i< mapnum; i++)
    {
        /* Create DataSlice */
        DataSlice arg_slice = DataSlice(tdata_arg_ptrs[i],
                                        sizes[i],
                                        static_cast<MapKind>(kinds[i]),
                                        debug_,
                                        perf_
                                        );
        DataSliceID arg_slice_id = arg_slice.id();
        MapKind arg_slice_kind = arg_slice.kind();
        /* If arg_slice was a unique slice, it would be present in 
         * either tdata_unique_slices_ or tmalloc_unique_slices_ */ 
        
        /* Check for presence in tdata_unique_slices_ */
        if (tdata_unique_slices_.count(arg_slice_id) > 0)
        {
            /* If MAP_FROM kind start a blocking read from its device buffer 
             * to make the device data consistent on the host */
            if (arg_slice_kind == MAP_FROM
                    || arg_slice_kind == MAP_TO_FROM)
            {
                tdata_unique_slices_[arg_slice_id].ReadFromDevice(queue_, true);
            }
            
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET_DATA_END: "
                          << "Removing section: " 
                          << std::endl << "\t"
                          << tdata_unique_slices_[arg_slice_id]
                          << std::endl;
            }
            /* Remove the slice from the unique slice list */
            tdata_unique_slices_.erase(arg_slice_id);
            continue;
        }

        /* NOTE: We dont check for tmalloc unique slices here 
         * These will be handled by GOMP_target_free */

        /* If arg_slice was a sub_slice, then it would be present in 
         * either tdata_sub_slices_ or tmalloc_sub_slices_ */ 
        
        /* Check if it is a tdata subslice */
        if (tdata_sub_slices_.count(arg_slice_id) > 0)
        {
            /* If MAP_FROM kind start a blocking read from its device buffer */
            if (arg_slice_kind == MAP_FROM
                    || arg_slice_kind == MAP_TO_FROM)
            {
                tdata_sub_slices_[arg_slice_id].ReadFromDevice(queue_, true);
            }
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET_DATA_END: "
                          << "Removing sub-section: " 
                          << std::endl << "\t"
                          << tdata_sub_slices_[arg_slice_id]
                          << std::endl;
            }
            /* Remove the sub slice from the sub slice list */
            tdata_sub_slices_.erase(arg_slice_id);
            continue;
        }

        /* Check if it is a tmalloc subslice */
        if (tmalloc_sub_slices_.count(arg_slice_id) > 0)
        {
            /* No need to CMEM_cacheInv this slice as it would have been 
             * done at the end of a target region */
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET_DATA_END: "
                          << "Removing sub-section: " 
                          << std::endl << "\t"
                          << tmalloc_sub_slices_[arg_slice_id]
                          << std::endl;
            }
            /* Remove the sub slice from the sub slice list */
            tmalloc_sub_slices_.erase(arg_slice_id);
        }

    }
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET_DATA_END: Close complete" 
            << std::endl;
    }
}

/*********************************************************************/
/* Initialize a single GOMP_Target regions specific variables        */
/*********************************************************************/
void Accelerator::InitializeTargetRegion( int mapnum,
                                          void **hostaddrs,
                                          unsigned int *sizes,
                                          unsigned char *kinds
                                        )
{
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Initializing region.."
                  << std::endl
                  << ">> TI_DEBUG: TARGET: Number of list items: " 
                  << mapnum 
                  << std::endl;
    }
    
    tregion_mapnum_ = mapnum;
    std::vector<void*> kernel_arg_ptrs; 
    ExtractArgPointers(hostaddrs, tregion_mapnum_, kernel_arg_ptrs);
    
    for (int i=0; i<tregion_mapnum_; i++)
    {
        tregion_slices_.push_back(DataSlice(kernel_arg_ptrs[i],
                                            sizes[i],
                                            static_cast<MapKind>(kinds[i]),
                                            debug_,
                                            perf_
                                            )
                                 );
    }
    
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Initialized data sections: "
                  << std::endl
                  << tregion_slices_ 
                  << std::endl;
        std::cout << ">> TI_DEBUG: TARGET: Region initialization complete" 
                  << std::endl;
    }
}

/*********************************************************************/
/* Extract pointers to kernel arguments and store for later use      */
/* Details: hostaddrs is originally a void* pointer and therefore 
 * cast to void*. It points to a struct containing pointers to 
 * kernel arguments. Since a void pointer cannot be dereferenced,
 * args is converted to an unsigned integer in args_int then cast to
 * char** struct pointer in the loop. Each dereferenced *struct_ptr
 * refers to a single pointer to a kernel argument which is then added
 * to the arg_ptr_list */
/*********************************************************************/
void Accelerator::ExtractArgPointers(void** hostaddrs,
                                     int arg_num,
                                     std::vector<void*>& arg_ptr_list)
{
    void *args = static_cast<void*>(hostaddrs);
    unsigned int args_int = reinterpret_cast<unsigned int>(args);
    unsigned int alignment = sizeof(void*);
    char **struct_ptr;
    void *buffer_ptr;
    for (int i=0; i<arg_num; i++)
    {
        struct_ptr = reinterpret_cast<char**>(args_int + i*alignment);
        buffer_ptr = reinterpret_cast<void*>(*struct_ptr);
        arg_ptr_list.push_back(buffer_ptr);
    }
}


/*********************************************************************/
/* Create a Map Buffer in given mem_region with given size and return*/
/* a pointer to that mapped region                                   */
/*********************************************************************/
void* Accelerator::CreateMapBuffer(__TI_omp_mem_region mem_region, int size)
{
    void *buffer_ptr = NULL;
    try
    {
        MapKind map_flag = MAP_TO_FROM;

        if(mem_region == TI_OMP_MEM_MSMC)
            map_flag = MAP_TO_MSMC;

        /* Create DataSlice */
        DataSlice map_slice (size, map_flag, debug_);
        /* Initialize device buffer */
        map_slice.InitDeviceBuffer(context_);
        
        buffer_ptr = queue_.enqueueMapBuffer(map_slice.device_buffer(), 
                                             CL_TRUE, 
                                             CL_MAP_READ|CL_MAP_WRITE, 
                                             0, 
                                             size
                                             );
        /* Set the start address of the slice */
        map_slice.set_start(buffer_ptr);
        DataSliceID map_slice_id = map_slice.id();
        /* Add (Copy) it to tmalloc unique slices */
        tmalloc_unique_slices_[map_slice_id] = map_slice;

        if(debug_)
        {
            std::cout << ">> TI_DEBUG: TI_OMP_DEVICE_ALLOC: "
                      << "Mapped section to device memory: "
                      << std::endl << "\t"
                      << map_slice 
                      << std::endl; 
        }    
    }
    catch (Error err) 
    { 
        std::cerr << ">> ERROR: Accelerator: CreateMapBuffer: " 
                  << err.what() << "(" << err.err() << ")" 
                  << std::endl; 
    }
    return buffer_ptr; 
}

/*********************************************************************/
/* Free the Map Buffer mapped to the given pointer                   */
/*********************************************************************/
void Accelerator::FreeMapBuffer(void* buffer_ptr)
{
    try
    {
        cl::Event ev;
        /* Get DataSliceID corresponding to the buffer_ptr start address */
        uintptr_t map_slice_start = reinterpret_cast<uintptr_t>(buffer_ptr);
        DataSliceID map_slice_id;
        bool id_present = false;
        for (auto& kv : tmalloc_unique_slices_)
        {
            /* Since tmalloc_unique_slices_ ID's are guaranteed to have u
             * nique start and end addresses, we can only find one 
             * candidate here */
            map_slice_id = kv.first;
            // TODO: Can we use Contains here?
            if (map_slice_id.start() == map_slice_start) 
            {
                id_present = true;
                break;
            }
        }
        if (!id_present)
        {
            std::cerr << ">> ERROR: TI_OMP_DEVICE_FREE: "
                      << "No corresponding section found"
                      << " for " << buffer_ptr
                      << std::endl;
            exit(EXIT_FAILURE);
        }
        /* Get reference to the corresponding data slice */ 
        DataSlice& map_slice = tmalloc_unique_slices_[map_slice_id];
        /* Get CL device buffer reference */
        cl::Buffer& map_slice_buffer = map_slice.device_buffer();

        queue_.enqueueUnmapMemObject(map_slice_buffer, 
                                     buffer_ptr, 
                                     NULL, 
                                     &ev
                                     );
        ev.wait();
        if (debug_)
        {
            std::cout << ">> TI_DEBUG: TI_OMP_DEVICE_FREE: Unmapped section: "
                      << std::endl << "\t"
                      << map_slice 
                      << std::endl;
        }

        /* Remove slice from tmalloc unique slices */
        tmalloc_unique_slices_.erase(map_slice_id);
    }
    catch (Error err) 
    { 
        std::cerr << ">> ERROR: Accelerator: FreeMapBuffer: " 
                  << err.what() << "(" << err.err() << ")" 
                  << std::endl; 
    }

}

/*********************************************************************/
/* Check for mapped buffers and flush them                           */
/*********************************************************************/
void Accelerator::InvAllMapBuffers(unsigned char *kinds)
{
    MapKind slice_kind;
    for (int i=0; i<tregion_mapnum_; i++)
    {
        /* Check if arg_ptr exists in tmalloc unique and sub slices */
        if (tregion_kinds_[i] == TI_OMP_DEVICE_ALLOC)
        {
            slice_kind = static_cast<MapKind>(kinds[i]);
            /* The buffer data must be invalidated
             * in order for host to have a consistent view of it 
             * only if the kind is MAP_FROM or MAP_TO_FROM */
            if (slice_kind == MAP_FROM || slice_kind == MAP_TO_FROM) 
            {
                CMEM_cacheInv(tregion_slices_[i].start(), 
                              tregion_slices_[i].size()
                              );
            }
        }
    }
}


/*********************************************************************/
/* Allocate input and output buffers to pass as kernel arguments     */
/*********************************************************************/
void Accelerator::CreateKernelArgs()
{
    bool slice_found;
    for (int i=0; i<tregion_mapnum_; i++)
    {
        slice_found = false;
        /* Get DataSlice reference */
        DataSlice& arg_slice = tregion_slices_[i];
        
        slice_found = CheckKernelArgInTmallocSlices(i, arg_slice); 
        
        if(slice_found)
            continue;
        
        slice_found = CheckKernelArgInTdataSlices(i, arg_slice);

        if(slice_found)
            continue;
        
        /*-------------------------------------------------------------------*/
        /* At this point, arg_slice is guaranteed to not be present in       */
        /* the tmalloc slices and tdata slices                               */
        /*-------------------------------------------------------------------*/
        
        CheckSliceOverlap(arg_slice, "TARGET");

        /*-------------------------------------------------------------------*/
        /* At this point, the arg_slice is guaranteed to not partially       */ 
        /* overlap any parent slice and is not contained within any of the   */
        /* parent slices i.e. it is a unique slice itself                    */
        /*-------------------------------------------------------------------*/
        
        /* Set TargetRegionKind */
        tregion_kinds_[i] = TARGET;
        
        /* Initialize its CL buffer */
        if (arg_slice.kind() == MAP_TO_L2)
        {
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET: Creating data section: " 
                          << std::endl << "\t"
                          << arg_slice 
                          << std::endl;
            }
            tregion_l2_sizes_[i] = arg_slice.size();
        }
        else 
        {
            if(debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET: Creating data section: " 
                          << std::endl << "\t"
                          << arg_slice 
                          << std::endl;
            }
            arg_slice.InitDeviceBuffer(context_); 
        }
    }
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Initialized all sections" 
                  << std::endl;
    }
}


/*********************************************************************/
/* Make the given DataSlice a tregion slice argument which will be   */
/* used to get the device_buffer for a CL kernel                     */
/*********************************************************************/
void Accelerator::MakeTregionSlice(int index,
                                   MapKind arg_slice_kind,
                                   DataSlice& slice, 
                                   TargetRegionKind slice_creator, 
                                   TargetRegionKind function_caller
                                   )
{ 
    if (slice_creator == TI_OMP_DEVICE_ALLOC)
    {
        /* If this is a TI_OMP_DEVICE_ALLOC slice:
         * If it is MAP_TO or MAP_TO_FROM
         * Write back host memory changes to device memory */
        if (arg_slice_kind == MAP_TO || arg_slice_kind == MAP_TO_FROM) 
            CMEM_cacheWb(slice.start(), slice.size());
    }
    else if (slice_creator == TARGET_DATA)
    {
        /* If this is a TARGET_DATA slice:
         * If it is MAP_TO or MAP_TO_FROM:
         * Wait for Host->Device transfer to complete */
        if (arg_slice_kind == MAP_TO || arg_slice_kind == MAP_TO_FROM) 
            slice.WaitOnWrite();
    }
    
    /* Set (Copy) the initialized slice */
    tregion_slices_[index] = slice;

    /* Set TargetRegionKind 
     * This represents the target region that this slice was created in */
    tregion_kinds_[index] = slice_creator;

    if (debug_)
    {
        std::cout << ">> TI_DEBUG: " << function_caller 
            << ": Found section allocated by "
            << slice_creator << ": "
            << std::endl << "\t"
            << slice
            << std::endl; 
    }
}


/*********************************************************************/
/* Check for presence of arg_slice in tmalloc slices and set the     */
/* tregion_slice at the given index if match is found                */
/*********************************************************************/
bool Accelerator::CheckKernelArgInTmallocSlices(int index, DataSlice& arg_slice)
{
    DataSliceID arg_slice_id = arg_slice.id();
    MapKind arg_slice_kind = arg_slice.kind();

    /* Check if arg_slice matches any slice in tmalloc_sub_slices_ */ 
    if (tmalloc_sub_slices_.count(arg_slice_id) > 0)
    {
        MakeTregionSlice(index, 
                         arg_slice_kind,
                         tmalloc_sub_slices_[arg_slice_id],
                         TI_OMP_DEVICE_ALLOC,
                         TARGET
                         );
        return true;
    }

    /* Check if arg_slice matches any slice in tmalloc_unique_slices_ */ 
    for (auto& kv : tmalloc_unique_slices_)
    {
        /* Here kv.second is the DataSlice in the pair kv */ 
        DataSlice& slice = kv.second;
        /* If start, size is the same, then use the same device buffer */ 
        if (slice == arg_slice)
        {
       
            MakeTregionSlice(index,
                             arg_slice_kind,
                             slice, 
                             TI_OMP_DEVICE_ALLOC,
                             TARGET
                             );
            return true;
        }
        else if (slice.Contains(arg_slice))
        {
            /* Write back host memory changes to device memory 
             * If the kind is MAP_TO or MAP_TO_FROM */
            if (arg_slice_kind == MAP_TO || arg_slice_kind == MAP_TO_FROM)
                CMEM_cacheWb(arg_slice.start(), arg_slice.size());

            /* arg_slice is *fully* contained in a unique slice, 
             * make it a sub slice of the unique slice */ 
            arg_slice.MakeSubSliceOf(slice);
            /* By making it a subslice the device buffer is also 
             * initialized */ 

            /* Set TargetRegionKind */
            tregion_kinds_[index] = TI_OMP_DEVICE_ALLOC;

            if (debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET: Found section "
                          << "allocated by TI_OMP_DEVICE_ALLOC: "
                          << std::endl << "\t"
                          << arg_slice 
                          << std::endl;
            }
            return true;
        }
    } 
    
    return false;
}

/*********************************************************************/
/* Check for presence of arg_slice in tdata slices and set the       */
/* tregion_slice at the given index if match is found                */
/*********************************************************************/
bool Accelerator::CheckKernelArgInTdataSlices(int index, DataSlice& arg_slice)
{
    DataSliceID arg_slice_id = arg_slice.id();
    MapKind arg_slice_kind = arg_slice.kind();
    
    /* Check if arg_slice matches any slice in tdata_sub_slices_ */ 
    if (tdata_sub_slices_.count(arg_slice_id) > 0)
    {
        
        MakeTregionSlice(index, 
                         arg_slice_kind,
                         tdata_sub_slices_[arg_slice_id],
                         TARGET_DATA,
                         TARGET
                         );
        return true;
    }

    /* Check if arg_slice matches any slice in tdata_unique_slices_ */ 
    for (auto& kv : tdata_unique_slices_)
    {
        /* Here kv.second is the DataSlice in pair kv */ 
        DataSlice& slice = kv.second;
        /* If start, size is the same, then use the same device buffer */ 
        if (slice == arg_slice)
        {
            MakeTregionSlice(index, 
                             arg_slice_kind,
                             slice,
                             TARGET_DATA,
                             TARGET
                             );

            return true;
        }
        else if (slice.Contains(arg_slice))
        {   
            /* Wait for host->device transfer to complete
             * If the kind is MAP_TO or MAP_TO_FROM */
            if (arg_slice_kind == MAP_TO || arg_slice_kind == MAP_TO_FROM) 
                slice.WaitOnWrite();
            /* arg_slice is *fully* contained in a unique slice, 
             * make it a sub slice of the unique slice */ 
            arg_slice.MakeSubSliceOf(slice);
            /* By making it a subslice the device buffer is also 
             * initialized */ 

            /* Set TargetRegionKind */
            tregion_kinds_[index] = TARGET_DATA;

            if (debug_)
            {
                std::cout << ">> TI_DEBUG: TARGET: Found section "
                          << "created by TARGET_DATA: "
                          << std::endl << "\t"
                          << arg_slice 
                          << std::endl;
            }
            return true;
        }
    }

    return false;
}

/*********************************************************************/
/* Read program binary from given char array and build it            */
/*********************************************************************/
void Accelerator::BuildProgramFromBinary(char bin_arr[], unsigned int bin_len)
{
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: Building program from binary" << std::endl;
        std::cout << ">> TI_DEBUG: Length of binary: " << bin_len << std::endl;
    }

    program_binary_ = cl::Program::Binaries(1, 
                                            std::make_pair(bin_arr, bin_len)
                                            );
    program_ = cl::Program(context_, 
                           devices_, 
                           program_binary_
                           );
    program_.build(devices_);      
    if (debug_)
        std::cout << ">> TI_DEBUG: Built program from binary" << std::endl;
}

/*********************************************************************/
/* Build the kernel with given function name and set its arguments   */
/*********************************************************************/
void Accelerator::BuildKernelAndSetArgs(std::string k_fn_name)
{
    /* Create Kernel */
    kernel_= cl::Kernel(program_, 
                        k_fn_name.c_str()
                        );   

    /* Set Kernel Arguments */
    for (int i=0; i<tregion_mapnum_; i++)
    {
        if (tregion_l2_sizes_.count(i) > 0)
            kernel_.setArg(i, 
                           __local(tregion_l2_sizes_[i])
                           );
        else
            kernel_.setArg(i, 
                           tregion_slices_[i].device_buffer()
                           );  
    }
    
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Ready to execute" 
                  << std::endl;
    }

}

/*********************************************************************/
/* Enqueue blocking write buffers                                    */
/*********************************************************************/
void Accelerator::EnqueueWriteBuffers()
{
    for (int i=0; i<tregion_mapnum_; i++)
    {   
        /* Check if this argument was only created in this 
         * target region */
        if (tregion_kinds_[i] == TARGET) 
        {
            DataSlice& slice = tregion_slices_[i];
            MapKind kind = slice.kind();
            if (kind == MAP_TO || kind == MAP_TO_FROM)
            { 
                /* Make host to device transfer */
                slice.WriteToDevice(queue_, false);
            } 
        } 
    }

    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Host->Device transfers initiated" 
                  << std::endl;
    }

}

/*********************************************************************/
/* Enqueue the kernel                                                */
/*********************************************************************/
void Accelerator::EnqueueKernel()
{
    /* If performance debug is enabled:
     * Finish writes now and report on their performance */
    if (perf_)
    {
        for (int i=0; i<tregion_mapnum_; i++)
        {   
            /* Check if this argument was only created in this 
             * target region */
            if (tregion_kinds_[i] == TARGET) 
            {
                DataSlice& slice = tregion_slices_[i];
                MapKind kind = slice.kind();
                if (kind == MAP_TO || kind == MAP_TO_FROM)
                { 
                    /* Finish write and report performance */
                    slice.WaitOnWrite();
                } 
            } 
        }
    } /* Else, writes will finish before kernel starts. 
         This is ensured as the queue_ is in-order */
    
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Executing.." 
                  << std::endl;
    }
    queue_.enqueueTask(kernel_, 
                       NULL, 
                       &k_event_
                       );
    k_event_.wait();
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Execution complete" 
                  << std::endl;
    }
}

/*********************************************************************/
/* Enqueue blocking read buffers                                     */
/*********************************************************************/
void Accelerator::EnqueueReadBuffers()
{
    for (int i=0; i<tregion_mapnum_; i++)
    {
        /* Check if this argument was only created in this 
         * target region */
        if (tregion_kinds_[i] == TARGET) 
        {

            if (tregion_slices_[i].kind() == MAP_FROM
                || tregion_slices_[i].kind() == MAP_TO_FROM
                )
            {
                /* Make device to host transfer */
                tregion_slices_[i].ReadFromDevice(queue_,true);
            } 
        }
    }
    queue_.finish();
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: Device->Host transfers complete" 
                  << std::endl;
    }

}

/*********************************************************************/
/* Report on kernel's peformance                                     */
/*********************************************************************/
void Accelerator::ReportKernelPerformance()
{
    if (perf_)
    {   
        cl_ulong t_start, t_end;
        k_event_.getProfilingInfo(CL_PROFILING_COMMAND_START,  &t_start);
        k_event_.getProfilingInfo(CL_PROFILING_COMMAND_END,    &t_end);

        /* Normalize the time to microseconds */
        t_start /= 1000; 
        t_end  /= 1000;

        std::cout << "\t> PERFORMANCE: Target region "
                  << "execution took " << t_end - t_start << "us"
                  << std::endl;
    }
}

/*********************************************************************/
/* Get ready for next GOMP_Target call                               */
/*********************************************************************/
void Accelerator::ClearTargetData()
{
    if (debug_)
    {
        std::cout << ">> TI_DEBUG: TARGET: End of region"
                  << std::endl;
    }
    tregion_slices_.clear();
    tregion_kinds_.clear();
    tregion_l2_sizes_.clear();
}

