/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file accelerator.h
 * \brief OpenCL Accelerator Wrapper  
 */

#ifndef ACCELERATOR_H_
#define ACCELERATOR_H_

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include <iostream>
#include <cstdio>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <sstream>
#include <iterator>
#include <unordered_map>
#include <vector>
#include <stack>
#include <ti/cmem.h>
#include "ocl_util.h"
#include "types.h"
#include "data_slice.h"

/**
 * \class Accelerator
 * \brief Wrapper class for an OpenCL Accelerator 
 * 
 * This class builds an OpenCL Accelerator instance and provides an API to 
 * manage initializing contexts, creating data buffers, queueing kernels
 * and reporting performance data.
 */
class Accelerator
{
    public:
        /**
         * \brief Constructor
         * \param debug Enable/disable printing of debugging information
         * \param type OpenCL device type
         */
        Accelerator(int type = CL_DEVICE_TYPE_ACCELERATOR);
        
        /**
         * \brief Destructor
         */
        ~Accelerator() {}
       
        /**
         * \brief Return CL initialization status 
         * \return true if initialized, false if not
         */
        bool init() { return init_; }

        /**
         * \brief Return accelerator device id
         * \return integer device id of accelerator
         */
        int device_id() { return device_id_; }

        /**
         * \brief Initialize OpenCL Accelerator device 
         * \param device_id Index number of OpenCL device being initialized 
         */
        void InitializeCL(int device_id = 0);
        
        /**
         * \brief Initialize the environment specific to a single 
         *        OpenMP target region
         * \param mapnum Number of data buffers being mapped to this region
         * \param hostaddrs Pointer to pointers to data buffers being mapped 
         * \param sizes Sizes of data buffers being mapped
         * \param kinds Kinds of data buffers being mapped i.e.
         *              Read/Write/Read+Write
         */
        void InitializeTargetRegion(int mapnum, 
                                    void **hostaddrs, 
                                    unsigned int *sizes, 
                                    unsigned char *kinds
                                    );
        /**
         * \brief Setup data items specific to a target data region
         * \param mapnum Number of data buffers being mapped to this region
         * \param hostaddrs Pointer to pointers to data buffers being mapped 
         * \param sizes Sizes of data buffers being mapped
         * \param kinds Kinds of data buffers being mapped i.e.
         *              Read/Write/Read+Write
         */
        void SetupTargetDataRegion(int mapnum,
                                   void **hostaddrs,
                                   unsigned int *sizes,
                                   unsigned char *kinds
                                   );

        /**
         * \brief Destroy data for a target data region
         * \param mapnum Number of data buffers being mapped to this region
         * \param hostaddrs Pointer to pointers to data buffers being mapped 
         * \param sizes Sizes of data buffers being mapped
         * \param kinds Kinds of data buffers being mapped i.e.
         *              Read/Write/Read+Write
         */
        void DestroyTargetDataRegion(int mapnum,
                                     void** hostaddrs,
                                     unsigned int* sizes,
                                     unsigned char* kinds
                                     );
        /**
         * \brief Make a target update for the given list items
         * \param mapnum Number of data buffers being mapped to this region
         * \param hostaddrs Pointer to pointers to data buffers being mapped 
         * \param sizes Sizes of data buffers being mapped
         * \param kinds Kinds of data buffers being mapped i.e.
         *              Read/Write/Read+Write
         */
        void MakeTargetUpdate(int mapnum,
                              void **hostaddrs,
                              unsigned int *sizes,
                              unsigned char *kinds
                              );

        /**
         * \brief Create a memory mapped buffer to a region of memory on the 
         *        target device
         * \param mem_region The region of target device memory on which to 
         *                   create the mapped buffer i.e. DDR or MSMC SRAM
         * \param size Size of memory to map
         * \return Host pointer to the mapped target memory
         */
        void* CreateMapBuffer(__TI_omp_mem_region mem_region, int size);
        
        /**
         * \brief Free a mapped a region of memory on the target device
         * \param buffer_ptr Host pointer to the mapped memory region 
         */
        void FreeMapBuffer(void* buffer_ptr);

         /**
          * \brief Check for mapped buffers and invalidate them for host
          *        consistency
          * \param kinds Kinds of data slices that were mapped i.e.
          *              Read/Write/Read+Write
          */
        void InvAllMapBuffers(unsigned char *kinds);
        
        /**
         * \brief Initialize data buffer arguments to the OpenCL Kernel 
         *        for a single target region
         */
        void CreateKernelArgs();

        /**
         * \brief Build OpenCL program from the embedded program binary in
         *        the __TI_CLACC_KERNEL_dsp_bin[] char array linked in by
         *        CLACC during creation of the program executable. 
         * \param bin_arr Char array containing OpenCL program binary
         * \param bin_len Length (in number of bytes) of the program binary
         * \note This is run once per lifetime of the program and not for 
         *       every target region
         */
        void BuildProgramFromBinary(char bin_arr[], unsigned int bin_len);
        
        /**
         * \brief Build OpenCl kernel for a single target region and set
         *        kernel arguments with the appropriate data buffers
         * \param k_fn_name Name of kernel function in OpenCL program
         */
        void BuildKernelAndSetArgs(std::string k_fn_name);
        
        /**
         * \brief Enqueue data buffers that are to be written to the 
         *        target device before enqueueing the kernel
         */
        void EnqueueWriteBuffers();
        
        /**
         * \brief Enqueue the OpenCL kernel for a single target region
         */
        void EnqueueKernel();
        
        /**
         * \brief Enqueue data buffers that are to be read from the 
         *        target device after execution of the kernel
         */
        void EnqueueReadBuffers();
        
        /**
         * \brief Report performance data for the kernel execution 
         *        from OpenCL command queue
         */
        void ReportKernelPerformance();
        
        /**
         * \brief Clear data specific to a single target region and get 
         *        acclerator instance ready to execute next target region
         */
        void ClearTargetData();

    private:
        bool                                    init_;
        bool                                    debug_;
        bool                                    perf_;
        int                                     device_id_;

        /*********************************************************************/
        /* OpenCL specific variables                                         */
        /*********************************************************************/
        cl::Context                             context_;
        cl::CommandQueue                        queue_;
        cl::Program::Binaries                   program_binary_;
        cl::Program                             program_;
        cl::Kernel                              kernel_;
        cl::Event                               k_event_;
        std::vector<cl::Device>                 devices_;
        long long                               timer_resolution_;
        cl_uint                                 base_addr_align_;

        /*********************************************************************/
        /* Target region specific variables                                  */
        /*********************************************************************/
        int                                     tregion_mapnum_;
        std::vector<DataSlice>                  tregion_slices_;
        std::unordered_map<int, int>            tregion_l2_sizes_;
        std::unordered_map<int, 
                           TargetRegionKind>    tregion_kinds_;
        
        /*********************************************************************/
        /* Target malloc maps                                                */
        /*********************************************************************/
        std::unordered_map<DataSliceID, 
                           DataSlice>           tmalloc_unique_slices_;
        std::unordered_map<DataSliceID, 
                           DataSlice>           tmalloc_sub_slices_;

        /*********************************************************************/
        /* Target data maps                                                  */
        /*********************************************************************/
        std::unordered_map<DataSliceID, 
                           DataSlice>           tdata_unique_slices_;
        std::unordered_map<DataSliceID, 
                           DataSlice>           tdata_sub_slices_;

        /**
         * \brief Extract pointers to data buffers to be mapped to a target 
         *        region
         * \param hostaddrs     Pointer to struct containing pointers to args
         * \param arg_num       Number of args to be extracted
         * \param arg_ptr_list  List to add arg pointers to
         */
        void ExtractArgPointers(void** hostaddrs,
                                int arg_num,
                                std::vector<void*>& arg_ptr_list
                                );

        
        /**
         * \brief Update slice if found in tdata slices 
         * \param arg_slice Reference to DataSlice
         * \return true if update performed, false otherwise
         */
        bool UpdateTdataSlices(DataSlice& arg_slice);
        
        /**
         * \brief Update slice if found in tmalloc slices
         * \param arg_slice Reference to DataSlice
         * \return true if update performed, false otherwise
         */
        bool UpdateTmallocSlices(DataSlice& arg_slice);
 
        /**
         * \brief Check presence of slice in given maps and handle if found      
         * \param arg_slice Reference to DataSlice
         * \param unique_slices Map of unique slices
         * \param sub_slices    Map of sub slices
         * \param region        Kind of target region the maps belong to
         * \return true if arg_slice found, false otherwise
         */
        bool CheckForSlice(DataSlice& arg_slice,
                           std::unordered_map<DataSliceID,DataSlice>& unique_slices,
                           std::unordered_map<DataSliceID,DataSlice>& sub_slices,
                           const std::string& region
                           );
        
        /**
         * \brief Check for overlap with tdata and tmalloc slices 
         *        and exit if found 
         * \param arg_slice Reference to DataSlice
         */
        void CheckSliceOverlap(DataSlice& arg_slice, std::string region);
        
        /**
         * \brief Make the given DataSlice a tregion slice argument which will be
         *        used to get the device_buffer for a CL kernel  
         * \param index             The index of tregion_slices that this slice 
         *                          will be assigned to
         * \param arg_slice_kind    The MapKind of the slice in the given target
         *                          region
         * \param slice             Reference to the DataSlice being handled
         * \param slice_creator     The kind of target region that created slice
         * \param function_caller   The kind of target region this function is 
         *                          being called from
         */
        void MakeTregionSlice(int index,
                              MapKind arg_slice_kind,
                              DataSlice& slice, 
                              TargetRegionKind slice_creator, 
                              TargetRegionKind function_caller
                              );

        /**
         * \brief Check for presence of arg_slice in tmalloc slices and set the 
         *        tregion_slice at the given index if match is found
         * \param arg_slice Reference to DataSlice
         * \return true if arg_slice found, false otherwise
         */
        bool CheckKernelArgInTmallocSlices(int index, DataSlice& arg_slice);
        
        /**
         * \brief Check for presence of arg_slice in tdata slices and set the 
         *        tregion_slice at the given index if match is found
         * \param arg_slice Reference to DataSlice
         * \return true if arg_slice found, false otherwise
         */
        bool CheckKernelArgInTdataSlices(int index, DataSlice& arg_slice);
        
};

/**
 * \brief Pretty print vectors to a stream
 * \param os Output stream reference 
 * \param v  Vector of any type
 * \return Reference to output stream object  
 */
template<class T>
std::ostream& operator<<(std::ostream& os, const std::vector<T>& v)
{
    for (unsigned int i = 0; i < v.size(); i++)
    {
        os << "\t" << v[i];
        if (i != v.size() - 1)
            os << std::endl;
    }
    return os;
}

#endif /* ACCELERATOR_H_ */
