/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions are met:
 *       * Redistributions of source code must retain the above copyright
 *         notice, this list of conditions and the following disclaimer.
 *       * Redistributions in binary form must reproduce the above copyright
 *         notice, this list of conditions and the following disclaimer in the
 *         documentation and/or other materials provided with the distribution.
 *       * Neither the name of Texas Instruments Incorporated nor the
 *         names of its contributors may be used to endorse or promote products
 *         derived from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file data_slice.cc
 * \brief Abstraction of a single slice of data used by an Accelerator 
 */

#include "data_slice.h"


/*********************************************************************/
/* Constructor: Setup data slice                                     */
/*********************************************************************/
DataSlice::DataSlice(unsigned int size, 
                     MapKind kind, 
                     bool debug, 
                     bool perf
                     )
{
    init_ = false;
    size_ = size;
    kind_ = kind;
    debug_ = debug;
    show_perf_ = perf;
    parent_id_ = DataSliceID(); 
}

/*********************************************************************/
/* Constructor: Setup data slice                                     */
/*********************************************************************/
DataSlice::DataSlice(void* start, 
                     unsigned int size, 
                     MapKind kind, 
                     bool debug, 
                     bool perf
                     )
{ 
    init_ = true;
    size_ = size;
    kind_ = kind;
    debug_ = debug;
    show_perf_ = perf;
    uintptr_t start_val = reinterpret_cast<uintptr_t>(start);
    uintptr_t end_val = start_val + size - 1; /* -1 as start_ is inclusive */
    id_ = DataSliceID(start_val, end_val);
    parent_id_ = DataSliceID(); 
}

/*********************************************************************/
/* Mutator: Set start address                                        */
/*********************************************************************/
void DataSlice::set_start(void* start)
{ 
    init_ = true;
    uintptr_t start_val = reinterpret_cast<uintptr_t>(start);
    uintptr_t end_val = start_val + size_ - 1; /* -1 as start_ is inclusive */
    id_ = DataSliceID(start_val, end_val);
}

/*********************************************************************/
/* Make this slice a subslice of the given parent DataSlice          */
/* i.e. create a CL sub-buffer out of parent's CL buffer             */
/*********************************************************************/
void DataSlice::MakeSubSliceOf(DataSlice& parent_slice)
{
    uintptr_t start_val        = id_.start();
    uintptr_t parent_start_val = parent_slice.id().start();

    /* Calculate start relative to parent slice */ 
    unsigned int region_start = static_cast<unsigned int>(start_val - parent_start_val);

    cl_buffer_region region = {region_start, size_};

    /* Get CL memory to initialize device buffer with 
     * Same flags as parent slice, ignore own kind */
    cl_mem_flags mem_flags = parent_slice.device_buffer()
                             .getInfo<CL_MEM_FLAGS>();
    
    /* Create CL SubBuffer */
    device_buffer_  = parent_slice.device_buffer()
                      .createSubBuffer(mem_flags,
                                       CL_BUFFER_CREATE_TYPE_REGION, 
                                       &region
                                       );
    /* Set parent ID */
    parent_id_ = parent_slice.id();

    if (init_ && debug_)
    {
        std::cout << ">> TI_DEBUG: DATA_SECTION: Created sub-section: " 
                  << std::endl << "\t"
                  << *this 
                  << std::endl << "\t"
                  << "from parent section: "
                  << std::endl << "\t"
                  << parent_slice
                  << std::endl; 
    }

}

/*********************************************************************/
/* Initialize the CL device buffer for this slice                    */
/*********************************************************************/
void DataSlice::InitDeviceBuffer(cl::Context& device_context)
{
    /* Get CL mem flags */
    cl_mem_flags mem_flags = GetCLMemFlags(kind_);

    /* Create device buffer */
    device_buffer_ =  cl::Buffer(device_context, 
                                 mem_flags, 
                                 size_
                                 );
    if (init_ && debug_)
    {
        std::cout << ">> TI_DEBUG: DATA_SECTION: "
                  << "Initialized device buffer for section: " 
                  << std::endl << "\t"
                  << *this 
                  << std::endl; 
    }
}

/*********************************************************************/
/* Get OpenCL memory flags associated with a MapKind                 */
/*********************************************************************/
cl_mem_flags DataSlice::GetCLMemFlags(MapKind kind)
{
    cl_mem_flags mem_flags = CL_MEM_READ_ONLY;
    
    switch (kind)
    {
        case MAP_TO:
            mem_flags = CL_MEM_READ_ONLY;
            break;
        case MAP_FROM:
            mem_flags = CL_MEM_WRITE_ONLY;
            break;
        case MAP_TO_FROM:
            mem_flags = CL_MEM_READ_WRITE; 
            break;
        case MAP_ALLOC:
            mem_flags = CL_MEM_READ_WRITE;
            break;
        case MAP_TO_MSMC:
            mem_flags = CL_MEM_READ_WRITE|CL_MEM_USE_MSMC_TI;
            break;
        case MAP_TO_L2: /* NOTE: Only here for completeness, not functional yet*/
            mem_flags = CL_MEM_READ_ONLY;
            break;
        default:
            std::cerr << ">> ERROR: DATA_SECTION: MapKind "
                      <<  kind << "not recognized" 
                      << std::endl;
            exit(EXIT_FAILURE);
    }
    return mem_flags;
}

/*********************************************************************/
/* Initiate transfer of host slice data to device buffer             */
/*********************************************************************/
void DataSlice::WriteToDevice(cl::CommandQueue& queue, bool blocking)
{
    cl_bool blocking_write = CL_TRUE;
    if(!blocking)
        blocking_write = CL_FALSE;

    queue.enqueueWriteBuffer(device_buffer_, /* CL Device buffer to write to */
            blocking_write, /* Synchronous or Asynchronous Write */
            0, /* Offset: Host index to start writing from */
            size_, /* Size of host memory to write */ 
            start(), /* Location of host memory to write from */ 
            NULL, /* Event wait list */
            &event_ /* Event to profile this write on */
            );
    
    if (init_ && debug_)
    {
        std::cout << ">> TI_DEBUG: DATA_SECTION: "
                  << "Transferred Host->Device: " 
                  << std::endl << "\t"
                  << *this 
                  << std::endl; 
    }

    if (init_ && blocking && show_perf_)
    {
        ShowPerformanceData("Host->Device");
    }
}

/*********************************************************************/
/* Initiate transfer of device buffer data to host                   */
/*********************************************************************/
void DataSlice::ReadFromDevice(cl::CommandQueue& queue, bool blocking)
{
    cl_bool blocking_read = CL_TRUE;
    if(!blocking)
        blocking_read = CL_FALSE;

    queue.enqueueReadBuffer(device_buffer_, /* CL Device buffer to read from */
            blocking_read, /* Synchronous or Asynchronous Read */
            0, /* Offset: Device index to start reading from */
            size_, /* Size of device memory to read */ 
            start(), /* Location of host memory to write to */ 
            NULL, /* Event wait list */
            &event_ /* Event to profile this read on */
            );
    
    if (init_ && debug_)
    {
        std::cout << ">> TI_DEBUG: DATA_SECTION: "
                  << "Transferred Device->Host: " 
                  << std::endl << "\t"
                  << *this 
                  << std::endl; 
    }
    
    if (init_ && blocking && show_perf_)
    {
        ShowPerformanceData("Device->Host");
    }

}

/*********************************************************************/
/* Wait for write event to complete                                  */
/*********************************************************************/
void DataSlice::WaitOnWrite()
{
    event_.wait();
    
    if (init_ && show_perf_)
    {   
        ShowPerformanceData("Host->Device");
    }
}

/*********************************************************************/
/* Wait for read event to complete                                  */
/*********************************************************************/
void DataSlice::WaitOnRead()
{
    event_.wait();
    
    if (init_ && show_perf_)
    {   
        ShowPerformanceData("Device->Host");
    }
}

/*********************************************************************/
/* Make a blocking write to device buffer or blocking read data from */
/* device buffer based on given kind on given queue                  */
/*********************************************************************/
void DataSlice::Update(cl::CommandQueue& queue, MapKind kind)
{
    if (init_ && debug_)
    {
        std::cout << ">> TI_DEBUG: DATA_SECTION: "
                  << "Performing update "
                  << kind 
                  << std::endl; 
    }

    if (kind == MAP_TO)
    {
        WriteToDevice(queue, true);
    }
    else if (kind == MAP_FROM)
    {
        ReadFromDevice(queue, true);
    }
}

/*********************************************************************/
/* Compare two data slices and return true if both slices have same  */
/* start, size                                                       */
/*********************************************************************/
bool DataSlice::operator==(const DataSlice& slice) const
{   
    if (id_ == slice.id_)
        return true;

    return false;
}


/*********************************************************************/
/* Get time elapsed for a read/write on this slice in microseconds   */
/*********************************************************************/
cl_ulong DataSlice::GetExecutionTime() const 
{
     cl_ulong t_start, t_end;

     event_.getProfilingInfo(CL_PROFILING_COMMAND_START,  &t_start);
     event_.getProfilingInfo(CL_PROFILING_COMMAND_END,    &t_end);

     /* Normalize the time to microseconds */
     t_start /= 1000; 
     t_end  /= 1000;

     return t_end-t_start;
}

/*********************************************************************/
/* Show read/write performance data                                  */
/*********************************************************************/
void DataSlice::ShowPerformanceData(std::string type)
{
    std::cout << "\t"
              << "> PERFORMANCE: Section " 
              << "[" << start() << "," << end() << "]: "
              << type << " transfer of " 
              << size_ << " bytes took "
              << GetExecutionTime() << "us"
              << std::endl;
}
