/******************************************************************************
 * Copyright (c) 2014, Texas Instruments Incorporated - http://www.ti.com/
 *   All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above copyright
 *        notice, this list of conditions and the following disclaimer in the
 *        documentation and/or other materials provided with the distribution.
 *      * Neither the name of Texas Instruments Incorporated nor the
 *        names of its contributors may be used to endorse or promote products
 *        derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *  THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

/**
 * \file data_slice_id.h
 * \brief Represents a [start_address, end_address] 
 */

#ifndef __DATA_SLICE_ID_H__
#define __DATA_SLICE_ID_H__

#include <iostream>
#include <iomanip>
#include <cstdint>
#include <utility>

/**
 * \class DataSliceID
 * \brief Identifier for a single DataSlice
 *
 * Unique identifier for a single DataSlice which can be used 
 * as a key in STL containers such as std::map and std::unordered_map
 */
class DataSliceID
{
    public:
        DataSliceID(uintptr_t start_addr = 0, uintptr_t end_addr = 0);
         
        //! Returns a combined unique hash value start, end
        uint64_t hash() const { return hash_combined_; }

        bool operator== (const DataSliceID& ID) const;
        bool operator<  (const DataSliceID& ID) const;

        bool Contains (uintptr_t address)     const;
        bool Contains (const DataSliceID& id) const;

        uintptr_t start() const { return id_.first; }
        uintptr_t end()   const { return id_.second; }

        friend std::ostream& operator<<(std::ostream&      os, 
                                        const DataSliceID& id);
    private:
        std::pair<uintptr_t, uintptr_t>  id_; // start address, end address
        uint64_t                         hash_combined_;
};

namespace std {
    /**
     * \brief Specialization for std::hash functor
     * \param ID DataSliceID reference
     * \return Hash value based on hashed ID components 
     *
     * Hash Functor for DataSliceID so it can be used as a key
     * in STL containers such as std::map, std::unordered_map 
     */
    template<>
        struct hash<DataSliceID>
        {
            uint64_t operator()(const DataSliceID& ID) const
            {
                return ID.hash();
            }
        };
}


//! Returns true if start, end are identical
inline bool DataSliceID::operator==(const DataSliceID& ID) const 
{
    if (start() == ID.start() && end() == ID.end())
        return true;

    return false;
}


//! Returns true if first component of is less than ID's first component
inline bool DataSliceID::operator<(const DataSliceID& ID) const 
{
    return (start() < ID.start());
}


/*********************************************************************/
/* Check if a data slice contains a memory address and return        */
/* true if the address falls within the slice borders                */
/*********************************************************************/
inline bool DataSliceID::Contains(uintptr_t address) const
{
    /* Check extremities */
    if (address < start() || address > end())
        return false;
    
    return true;
}

/*********************************************************************/
/* Check if a data slice contains given data slice and return true   */ 
/* if the given slice falls completely within the slice borders      */
/*********************************************************************/
inline bool DataSliceID::Contains(const DataSliceID& id) const
{
    /* Check if given slice start and end is in range 
     * If either of them are not in range, then the slice is not 
     * completely in range, return false */
    if (!Contains(id.start()) || !Contains(id.end()))
        return false;

    /* If both slice start and end are in range, then the slice is in range */
    return true;
}

#endif /* __DATA_SLICE_ID_H__ */
